# distutils: language = c
# cython: boundscheck = False
# cython: wraparound = False
# cython: profile = False

import numpy as np
cimport numpy as np
cimport cython
from cython.parallel import prange
from libc.stdlib cimport malloc, free

ctypedef np.int32_t NP_INT_t
ctypedef np.float64_t NP_FLOAT_t

cdef extern from "<math.h>" nogil:
    const float INFINITY


cdef extern from "<limits.h>":
    const int INT_MIN
    const int INT_MAX


cdef struct IMM_Cut:
    int col
    float threshold


@cython.boundscheck(False)
@cython.wraparound(False)
def get_min_mistakes_cut(NP_FLOAT_t[:,:] X, NP_INT_t[:] y, NP_FLOAT_t[:,:] centers, NP_INT_t[:] valid_centers, NP_INT_t[:] valid_cols, int njobs):
    cdef int n = X.shape[0]
    cdef int k = centers.shape[0]
    cdef int d = valid_cols.shape[0]
    cdef int *centers_count = <int *> malloc(k * sizeof(int))
    cdef NP_FLOAT_t *cols_thresholds = <NP_FLOAT_t *> malloc(d * sizeof(NP_FLOAT_t))
    cdef int *cols_mistakes = <int *> malloc(d * sizeof(int))
    cdef int col
    cdef int best_col = -1
    cdef NP_FLOAT_t best_threshold
    cdef int min_mistakes = INT_MAX

    # Count the number of data points for each center.
    # This information will be helpful for fast mistakes calculation, once a threshold pass a center.
    for i in range(k):
        centers_count[i] = 0
    for i in range(n):
        centers_count[y[i]] += 1

    if njobs is None or njobs <= 1:
        # Iterate over valid coordinates
        for col in range(d):
            if valid_cols[col] == 1:
                update_col_min_mistakes_cut(X, y, centers, valid_centers, centers_count, cols_thresholds, cols_mistakes, col, n, d, k)
    else:
        # Iterate over valid coordinates
        for col in prange(d, nogil=True, num_threads=njobs):
            if valid_cols[col] == 1:
                update_col_min_mistakes_cut(X, y, centers, valid_centers, centers_count, cols_thresholds, cols_mistakes, col, n, d, k)

    for col in range(d):
        # This is a valid column
        if valid_cols[col] == 1:
            # We found a valid split
            if cols_mistakes[col] != -1:
                # This is a better cut
                if cols_mistakes[col] < min_mistakes:
                    best_col = col
                    min_mistakes = cols_mistakes[col]

    if best_col != -1:
        best_threshold = cols_thresholds[best_col]

    free(cols_thresholds)
    free(cols_mistakes)
    free(centers_count)

    if best_col == -1:
        return None
    else:
        return IMM_Cut(best_col, best_threshold)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void update_col_min_mistakes_cut(NP_FLOAT_t[:,:] X, NP_INT_t[:] y, NP_FLOAT_t[:,:] centers, NP_INT_t[:] valid_centers, int* centers_count, NP_FLOAT_t *cols_thresholds, int *cols_mistakes, int col, int n, int d, int k) nogil:
    cdef int i
    cdef int ix
    cdef int ic
    cdef int mistakes
    cdef NP_FLOAT_t prev_threshold
    cdef NP_FLOAT_t threshold
    cdef NP_FLOAT_t max_val
    cdef np.int64_t[:] data_order
    cdef np.int64_t[:] centers_order
    cdef int *left_centers_count = <int *> malloc(k * sizeof(int))
    cdef int curr_center_idx
    cdef bint valid_found = 0
    cdef bint is_center_threshold
    cdef int min_mistakes = INT_MAX
    cdef NP_FLOAT_t best_threshold

    # Sort data points and centers
    with gil:
        data_order = np.asarray(X[:,col]).argsort()
        centers_order = np.asarray(centers[:, col]).argsort()

    # Find maximal value of valid centers. Possible threshold must be strictly smaller than that.
    max_val = -INFINITY
    for i in range(k):
        if valid_centers[i] == 1:
            if centers[i, col] > max_val:
                max_val = centers[i, col]

    # For each center, count number of data points associated with it that are smaller than the current threshold.
    # This information will be helpful for fast mistakes calculation, once a threshold pass a center.
    for i in range(k):
        left_centers_count[i] = 0

    # Initialize pointers to data points and centers sorted lists.
    ix = 0
    ic = 0

    # Initialize number of mistakes.
    mistakes = 0

    # Advance center index to the first valid one
    while valid_centers[centers_order[ic]] == 0:
        ic += 1

    # The first threshold is the value of the first valid center.
    threshold = centers[centers_order[ic], col]

    # The first threshold is a center.
    # All data points that are smaller or equal to the threshold will be accumulated prior to the main loop.
    is_center_threshold = 1

    # Advance data point index to the first valid center (which is the first valid threshold).
    # Count mistakes of points smaller than the first valid threshold
    while ix < n and X[data_order[ix], col] <= threshold:
        curr_center_idx = y[data_order[ix]] # Center of the current data point
        left_centers_count[curr_center_idx] += 1
        if centers[curr_center_idx, col] >= threshold:
            mistakes += 1
        ix += 1

    # Corner case.
    # Exactly n - 1 data points are on the left of the first center - this is a valid cut, but we won't enter the main loop.
    if ix == n - 1:
        # Recalculate the number of mistakes.
        # In this corner case all points except one are to the left of the current threshold.
        mistakes = 0
        ic = 0
        # Go over all valid centers.
        while ic < k:
            if valid_centers[ic] != 0:
                # If a center is to the right of the current threshold, then all of its point are considered as mistakes.
                # (Perhaps except to the last point that will be corrected later).
                if centers[ic, col] > threshold:
                    mistakes += centers_count[ic]
            ic += 1
        # Find the center of the single point that is on the right of the threshold.
        # If the center is also to the right of the current threshold, then remove one mistake.
        ic = y[data_order[n - 1]]
        if centers[ic, col] > threshold:
            mistakes -= 1
        # Update best cut.
        if mistakes < min_mistakes:
            valid_found = 1
            best_col = col
            best_threshold = threshold
            min_mistakes = mistakes

    # Main loop
    while ix < n - 1 and ic < k:

        # If threshold reached to the last valid center, the loop should end.
        if threshold >= max_val:
            break

        # In case this threshold is associated to a data point
        if is_center_threshold == 0:
            # Find data point center
            curr_center_idx = y[data_order[ix]]
            # Increase the count of points smaller than the threshold associated with the center
            left_centers_count[curr_center_idx] += 1

            # Update the mistakes count
            if centers[curr_center_idx, col] >= threshold:
                mistakes += 1
            elif centers[curr_center_idx, col] < threshold:
                mistakes -= 1

            # Move to the next data point index
            ix += 1

        # In case this threshold is associated to a center
        else:
            # Update mistakes count
            # left points are no longer mistakes
            # right points (which equal to total points - left points) are now mistakes
            mistakes += centers_count[centers_order[ic]] - 2 * left_centers_count[centers_order[ic]]

            # Move to the next center index
            ic += 1
            while valid_centers[centers_order[ic]] == 0 and ic < k:
                ic += 1

        prev_threshold = threshold

        # Find next threshold
        # in case of equality, data points arrive before centers in order to correctly find left count
        if X[data_order[ix], col] <= centers[centers_order[ic], col]:
            threshold = X[data_order[ix], col]
            is_center_threshold = 0
        else:
            threshold = centers[centers_order[ic], col]
            is_center_threshold = 1

        # Update best cut (only if the next threshold is not equal to the current one)
        if (prev_threshold != threshold) and (mistakes < min_mistakes):
            valid_found = 1
            best_threshold = prev_threshold
            min_mistakes = mistakes

    free(left_centers_count)

    if valid_found == 1:
        cols_thresholds[col] = best_threshold
        cols_mistakes[col] = min_mistakes
    else:
        cols_thresholds[col] = -1.0
        cols_mistakes[col] = -1


cdef struct Surrogate_Cut:
    int col
    float threshold
    NP_FLOAT_t cost
    int center_left
    int center_right


@cython.boundscheck(False)
@cython.wraparound(False)
def get_min_surrogate_cut(NP_FLOAT_t[:,:] X, NP_FLOAT_t[:,:] X_center_dot, NP_FLOAT_t[:] X_sum_all_center_dot, NP_FLOAT_t[:] centers_norm_sqr, int njobs):
    cdef int n = X.shape[0]
    cdef int k = X_center_dot.shape[1]
    cdef int d = X.shape[1]
    cdef int col
    cdef NP_FLOAT_t *thresholds = <NP_FLOAT_t *> malloc(d * sizeof(NP_FLOAT_t))
    cdef NP_FLOAT_t *costs = <NP_FLOAT_t *> malloc(d * sizeof(NP_FLOAT_t))
    cdef int *left_centers = <int *> malloc(d * sizeof(int))
    cdef int *right_centers = <int *> malloc(d * sizeof(int))
    cdef (NP_FLOAT_t, NP_FLOAT_t, int, int) cut
    cdef int best_col = -1
    cdef NP_FLOAT_t best_cost = INFINITY
    cdef NP_FLOAT_t best_threshold
    cdef int best_center_left
    cdef int best_center_right


    if njobs is None or njobs <= 1:
        # Iterate over valid coordinates
        for col in range(d):
            update_col_surrogate_cut(X, X_center_dot, centers_norm_sqr, X_sum_all_center_dot, n, d, k, col, thresholds, costs, left_centers, right_centers)
    else:
        # Iterate over valid coordinates
        for col in prange(d, nogil=True, num_threads=njobs):
            update_col_surrogate_cut(X, X_center_dot, centers_norm_sqr, X_sum_all_center_dot, n, d, k, col, thresholds, costs, left_centers, right_centers)

    for col in range(d):
        # This is a valid cut
        if left_centers[col] != -1:
            # This is a better cut
            if costs[col] < best_cost:
                best_col = col
                best_cost = costs[col]


    if best_col != -1:
        best_threshold = thresholds[best_col]
        best_center_left = left_centers[best_col]
        best_center_right = right_centers[best_col]

    free(thresholds)
    free(costs)
    free(left_centers)
    free(right_centers)

    if best_col == -1:
        return None
    else:
        return Surrogate_Cut(best_col, best_threshold, best_cost, best_center_left, best_center_right)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void update_col_surrogate_cut(NP_FLOAT_t[:,:] X, NP_FLOAT_t[:,:] X_center_dot, NP_FLOAT_t[:] centers_norm_sqr, NP_FLOAT_t[:] X_sum_all_center_dot, int n, int d, int k, int col, NP_FLOAT_t *thresholds, NP_FLOAT_t *costs, int *left_centers, int *right_centers) nogil:
    cdef int i
    cdef int ix
    cdef int ic
    cdef int n_left
    cdef int n_right
    cdef NP_FLOAT_t[:] curr_X_center_dot
    cdef NP_FLOAT_t *X_sum_left_center_dot = <NP_FLOAT_t *> malloc(k * sizeof(NP_FLOAT_t))
    cdef NP_FLOAT_t *X_sum_right_center_dot = <NP_FLOAT_t *> malloc(k * sizeof(NP_FLOAT_t))
    cdef NP_FLOAT_t cost
    cdef NP_FLOAT_t left_cost
    cdef NP_FLOAT_t right_cost
    cdef int left_center
    cdef int right_center
    cdef NP_FLOAT_t cur_total_cost
    cdef NP_FLOAT_t prev_threshold
    cdef NP_FLOAT_t threshold
    cdef np.int64_t[:] data_order
    cdef bint valid_found = 0
    cdef NP_FLOAT_t best_cost = INFINITY
    cdef NP_FLOAT_t best_threshold
    cdef int best_center_left
    cdef int best_center_right

    # Sort data points
    with gil:
        data_order = np.asarray(X[:,col]).argsort()

    ix = 0

    curr_X_center_dot = X_center_dot[data_order[0]]
    for i in range(k):
        X_sum_left_center_dot[i] = curr_X_center_dot[i]
        X_sum_right_center_dot[i] = X_sum_all_center_dot[i]
    for i in range(k):
        X_sum_right_center_dot[i] -= curr_X_center_dot[i]
    n_left = 1
    n_right = n - 1

    threshold = X[data_order[0], col]

    while ix < n - 1:

        ix += 1

        prev_threshold = threshold
        threshold = X[data_order[ix], col]

        if prev_threshold != threshold:

            left_cost = INFINITY
            right_cost = INFINITY
            for ic in range(k):
                cost = n_left * centers_norm_sqr[ic] - 2 * X_sum_left_center_dot[ic]
                if cost < left_cost:
                    left_cost = cost
                    left_center = ic
                cost = n_right * centers_norm_sqr[ic] - 2 * X_sum_right_center_dot[ic]
                if cost < right_cost:
                    right_cost = cost
                    right_center = ic

            cur_total_cost = left_cost +  right_cost

        # Add dot product of current vector and each center to the left
        # Subtract dot product of current vector and each center to the right
        curr_X_center_dot = X_center_dot[data_order[ix]]
        for i in range(k):
            X_sum_left_center_dot[i] = X_sum_left_center_dot[i] + curr_X_center_dot[i]
            X_sum_right_center_dot[i] = X_sum_right_center_dot[i] - curr_X_center_dot[i]

        n_left += 1
        n_right -= 1

        if (prev_threshold != threshold) and (cur_total_cost < best_cost):
            valid_found = 1
            best_threshold = prev_threshold
            best_cost = cur_total_cost
            best_center_left = left_center
            best_center_right = right_center


    free(X_sum_left_center_dot)
    free(X_sum_right_center_dot)

    if valid_found == 1:
        thresholds[col] = best_threshold
        costs[col] = best_cost
        left_centers[col] = best_center_left
        right_centers[col] = best_center_right
    else:
        thresholds[col] = -1.0
        costs[col] = -1.0
        left_centers[col] = -1
        right_centers[col] = -1


from typing import Union, Callable
from .explainer import MercuryExplainer
from ..explanations.shuffle_importance import FeatureImportanceExplanation


import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler


class ShuffleImportanceExplainer(MercuryExplainer):
    """
    This explainer estimates the feature importance of each predictor for a
    given black-box model. The used strategy consists on random shuffling one
    variable at a time and, on each step, checking how much a particular
    metric worses. The features which make the model to perform the worst are
    the most important ones.

    Args:
        eval_fn (Callable):
            Custom evaluation function. It will recieve a DataFrame with features and
            a Numpy array with the target outputs for each instance. It must
            implement an inference process and return a metric to score the model
            performance on the given data. This metric must be real numbered.
            If we use a metric which higher values means better metric (like accuracy)
            and we use the parameter `normalize=True` (default option), then it is
            recommended to return the negative of that metric in `eval_fn` to make
            the results more intuitive.
            In the case of Pyspark explanations, the function will only recieve
            a PySpark in the first argument already containing the target column,
            whereas the second argument will be None.
        normalize (bool):
            Whether to scale the feature importances between 0 and 1. If True, then
            it shows the relative importance of the features.
            If False, then the feature importances will be the value of the metric
            returned in `eval_fn` when shuffling the features.
            Default value is True

    Example:
        ```python
        # "Plain python" example
        >>> features = pd.read_csv(PATH_DATA)
        >>> targets = features['target']  # Targets
        >>> features = features.loc[:, FEATURE_NAMES] # DataFrame with only features
        >>> def my_inference_function(features, targets):
        ...     predictions = model.predict(features)
        ...     return mean_squared_error(targets, predictions)
        >>> explainer = ShuffleImportanceExplainer(my_inference_function)
        >>> explanation = explainer.explain(features, targets)
        >>> explanation.plot()

        # Explain a pyspark model (or pipeline)
        >>> features = sess.createDataFrame(pandas_dataframe)
        >>> target_colname = "target"  # Column name with the ground truth labels
        >>> def my_inference_function(features, targets):
        ...     model_inp = vectorAssembler.transform(features)
        ...     model_out = my_pyspark_transformer.transform(model_inp)
        ...     return my_evaluator.evaluate(model_out)
        >>> explainer = ShuffleImportanceExplainer(my_inference_function)
        >>> explanation = explainer.explain(features, target_colname)
        >>> explanation.plot()
        ```
    """
    def __init__(self,
                 eval_fn: Callable[[Union["pd.DataFrame", "pyspark.sql.DataFrame"], Union["np.ndarray", str]], float],  # noqa: F821
                 normalize: bool = True
        ):
        self.eval_fn = eval_fn
        self.normalize = normalize

    def explain(self,
                predictors: Union["pd.DataFrame", "pyspark.sql.DataFrame"],  # noqa: F821
                target: Union["np.ndarray", str]
        ) -> FeatureImportanceExplanation:
        """
        Explains the model given the data.

        Args:
            predictors (Union[pandas.DataFrame, pyspark.sql.DataFrame]):
                DataFrame with the features the model needs and that will be explained.
                In the case of PySpark, this dataframe must  also contain a column
                with the target.
            target (Union[numpy.ndarray, str]):
                The ground-truth target for each one of the instances. In the case of
                Pyspark, this should be the name of the column in the DataFrame which
                holds the target.

        Raises:
            ValueError: if type(predictors) == pyspark.sql.DataFrame && type(target) != str
            ValueError: if type(predictors) == pyspark.sql.DataFrame && target not in predictors.columns

        Returns:
            FeatureImportanceExplanation with the performances of the model
        """

        implementation = self.__impl_base
        feature_names = []
        # Cheap way of check if type(predictors) == pyspark.sql.DataFrame (without importing pyspark).
        if hasattr(type(predictors), 'toPandas'):
            if type(target) != str:
                raise ValueError("""If predictors is a Spark DataFrame, target should be the name \
                                 of the tareget column (a str)""")
            implementation = self.__impl_pyspark
            feature_names = list(filter(lambda x: x!=target, predictors.columns))
            if len(feature_names) == len(predictors.columns):
                raise ValueError(f"""`target` must be the name of the target column in the DataFrame. \
                                 Value passed: {target}""")
        else:
            feature_names = list(predictors.columns)
            if type(target) == str:
                feature_names = list(filter(lambda x: x!=target, feature_names))

        metrics = {}
        for col in feature_names:
            metrics[col] = implementation(predictors, target, col)
        if self.normalize:
            metrics = self._normalize_importances(metrics)
        return FeatureImportanceExplanation(metrics)

    def __impl_base(self, predictors, target, column):
        temp = predictors.copy()
        temp[column] = np.array(temp[column].sample(frac=1))
        return self.eval_fn(temp, target)

    def __impl_pyspark(self, predictors, target, column):
        from pyspark.sql.functions import rand, row_number, monotonically_increasing_id
        from pyspark.sql.window import Window

        # Shuffle column values creating a temporal rand column and ordering by it.
        # Then, we merge the ordered DF with the old one removing the original and
        # temporal columns
        shuffled = predictors.select(column).withColumn('rand', rand()).orderBy('rand')
        # In order to merge back we also need to include a row number.
        shuffled=shuffled.withColumn('shuff_id_idx', row_number().over(Window.orderBy(monotonically_increasing_id())))
        temp = predictors.withColumn('shuff_id_idx', row_number().over(Window.orderBy(monotonically_increasing_id())))

        temp = temp.drop(column).join(shuffled, on='shuff_id_idx').drop('shuff_id_idx').drop('rand')

        return self.eval_fn(temp, None)

    def _normalize_importances(self, metrics):
        df_metrics = pd.DataFrame.from_dict(metrics, orient="index", columns=["importance"])
        df_metrics["importance"] = MinMaxScaler().fit_transform(df_metrics["importance"].values.reshape(-1,1))
        return df_metrics["importance"].to_dict()



import numpy as np
import matplotlib.pyplot as plt
import typing as TP

from math import ceil

class PartialDependenceExplanation():
    """
    This class holds the result of a Partial Dependence explanation and
    provides functionality for plotting those results via Partial Dependence Plots.

    Args:
        data (dict):
            Contains the result of the PartialDependenceExplainer. It must be in the
            form of: ::
                {
                    'feature_name': {'values': [...], 'preds': [...], 'lower_quantile': [...], 'upper_quantile': [...]},
                    'feature_name2': {'values': [...], 'preds': [...], 'lower_quantile': [...], 'upper_quantile': [...]},
                    ...
                }

    """
    def __init__(self, data):
        self.data = data

    def plot_single(self, var_name: str, ax=None, quantiles:TP.Union[bool, list] = False, filter_classes:list = None, **kwargs):
        """
        Plots the partial dependence of a single variable.

        Args:
            var_name (str):
                Name of the desired variable to plot.
            quantiles (bool or list[bool]):
                Whether to also plot the quantiles and a shaded area between them. Useful to check whether the predictions
                have high or low dispersion. If data doesn't contain the quantiles this parameter will be ignored.
            filter_clases (list):
                List of bool with the classes to plot. If None, all classes will be plotted. Ignored if the target variable
                is not categorical.
            ax (matplotlib.axes._subplots.AxesSubplot):
                Axes object on which the data will be plotted.
        """
        # If user pass a single bool and prediction data is a multinomial, we conver the
        # single boolean to a mask array to only plot the quantile range over the selected
        # classes.
        if len(self.data[var_name]['preds'].shape)>=2:
            if type(quantiles) == list and len(quantiles) != self.data[var_name]['preds'].shape[1]:
                raise ValueError("len(quantiles) must be equal to the number of classes.")
            if type(quantiles) == bool:
                quantiles = [quantiles for i in range(self.data[var_name]['preds'].shape[1])]
        elif type(quantiles) == list and len(self.data[var_name]['preds'].shape)==1:
            quantiles = quantiles[0]

        if filter_classes is not None:
            filter_classes = np.where(filter_classes)[0].tolist()
        else:
            filter_classes = np.arange(self.data[var_name]['preds'].shape[-1]).tolist()
            if len(self.data[var_name]['preds'].shape) < 2:
                filter_classes = None

        ax = ax if ax else plt.gca()

        ax.set_title(var_name)
        ax.set_xlabel(f"{var_name} value")
        ax.set_ylabel("Avg model prediction")

        vals = np.array(self.data[var_name]['values'])
        int_locations = np.arange(len(vals))

        non_numerical_values = False
        # Check if variable is categorical. If so, plot bars
        if self.data[var_name]['categorical'] and not type(vals[0]) == float:
            bar_width = .2
            class_nb = 0 if not filter_classes else len(filter_classes)

            if type(vals[0]) == float or type(vals[0]) == int:
                ax.set_xticks(self.data[var_name]['values'])
            else:
                non_numerical_values = True
                bar_offsets = np.linspace(-bar_width, bar_width, num=class_nb) / class_nb
                ax.set_xticks(int_locations)
                ax.set_xticklabels(self.data[var_name]['values'])

            if class_nb == 0:
                # If prediction is a single scalar
                if non_numerical_values:
                    ax.bar(int_locations, self.data[var_name]['preds'], width=bar_width, label='Prediction',**kwargs)
                else:
                    ax.bar(vals, self.data[var_name]['preds'], width=bar_width, label='Prediction', **kwargs)

                if quantiles:
                    ax.errorbar(
                        int_locations,
                        self.data[var_name]['preds'],
                        yerr=np.vstack([self.data[var_name]['lower_quantile'],
                                        self.data[var_name]['upper_quantile']]),
                        fmt='ko',
                        label='Quantiles',
                        **kwargs
                    )

            else:
                # If prediction is multiclass
                for i in range(class_nb):
                    if i in filter_classes:
                        if non_numerical_values:
                            ax.bar(int_locations + bar_offsets[i], self.data[var_name]['preds'][:,i],
                                    width=bar_width / class_nb, label=f'Class {i}',**kwargs)
                        else:
                            ax.bar(vals, self.data[var_name]['preds'][:,i], width=bar_width / class_nb, label=f'Class {i}', **kwargs)

                    if quantiles[i]:
                        ax.errorbar(
                            int_locations + bar_offsets[i],
                            self.data[var_name]['preds'][:, i],
                            yerr=np.vstack([self.data[var_name]['lower_quantile'][:,i],
                                            self.data[var_name]['upper_quantile'][:,i]]),
                            fmt='ko',
                            label=f'Quantiles {i}',
                            **kwargs
                        )

            if class_nb > 0:
                ax.legend()

        else:  # Variable is continuous

            # Check whether prediction data is multinomial
            if filter_classes:
                objs = ax.plot(vals, self.data[var_name]['preds'][:, filter_classes], **kwargs)
            else:
                objs = ax.plot(vals, self.data[var_name]['preds'], **kwargs)
            if len(self.data[var_name]['preds'].shape)>=2:
                labels = [f"Class: {i}" for i in range(self.data[var_name]['preds'].shape[1])]
                # Filter labels
                labels = [l for i, l in enumerate(labels) if i in filter_classes]
                # Show labels
                ax.legend(iter(objs), labels)
                for i in range(self.data[var_name]['preds'].shape[1]):
                    if quantiles[i] and len(self.data[var_name]['lower_quantile']) > 0:
                        # Plot quantiles and a shaded band between them

                        # We will need the color assigned to each one of the lines so the
                        # shaded area also has that color. Since filtering can be done, we
                        # extract the line index as the minimum between the current class
                        # index and the maximum amount of lines on the canvas.
                        obj_index = min(i, len(objs) - 1)

                        # Actually plot the shaded area
                        ax.plot(vals, self.data[var_name]['lower_quantile'][:,i], ls='--', color=objs[obj_index].get_color(),**kwargs)
                        ax.plot(vals, self.data[var_name]['upper_quantile'][:,i], ls='--', color=objs[obj_index].get_color(), **kwargs)
                        ax.fill_between(vals,
                                self.data[var_name]['lower_quantile'][:,i], self.data[var_name]['upper_quantile'][:,i], alpha=.05)
            else:  # If target is not multinomial
                if quantiles and len(self.data[var_name]['lower_quantile']) > 0:
                    # Plot quantiles and a shaded band between them
                    ax.plot(vals, self.data[var_name]['lower_quantile'], ls='--', color=objs[0].get_color(),**kwargs)
                    ax.plot(vals, self.data[var_name]['upper_quantile'], ls='--', color=objs[0].get_color(), **kwargs)
                    ax.fill_between(vals, self.data[var_name]['lower_quantile'], self.data[var_name]['upper_quantile'], alpha=.05)

    def plot(self, ncols:int = 1, figsize:tuple = (15,15), quantiles:TP.Union[bool, list] = False, filter_classes:list = None, **kwargs):
        """
        Plots a summary of all the partial dependences.

        Args:
            ncols (int):
                Number of columns of the summary. 1 as default.
            quantiles (bool or list):
                Whether to also plot the quantiles and a shaded area between them. Useful to check whether the predictions
                have high or low dispersion. If this is a list of booleans, quantiles
                will be plotted filtered by class (i.e. `quantiles[0]` = `class number 0`).
            filter_clases (list):
                List of bool with the classes to plot. If None, all classes will be plotted. Ignored if the target variable
                is not categorical.
            figsize (tuple):
                Size of the plotted figure
        """
        features = list(self.data.keys())

        fig, ax = plt.subplots(ceil(len(features) / ncols), ncols, figsize=figsize)

        for i, feat_name in enumerate(features):
            sbplt = ax[i] if ncols==1 or ncols==len(features) else ax[i // ncols, i % ncols]
            self.plot_single(feat_name, sbplt, quantiles=quantiles, filter_classes=filter_classes, **kwargs)

    def __getitem__(self, key:str):
        """
        Gets the dependence data of the desired feature.

        Args:
            key (str):
                Name of the feature.
        """
        return self.data[key]['values'], self.data[key]['preds']



# This module contains dummy versions of Explainers with Alibi dependencies.
# This classes are imported in the __init__.py if the import of the original class fails
# If the user tries to import any of the classes, and error is raised indicating that alibi
# must be installed first

class _DummyAlibiExplainer:
    """
    Class which raises and error if instantiated
    """
    def __init__(self):
        raise ModuleNotFoundError("You need to install alibi library to use this explainer.")

class ALEExplainer(_DummyAlibiExplainer):

    def __init__(self, predictor, target_names):
        super().__init__()

class AnchorsWithImportanceExplainer(_DummyAlibiExplainer):

    def __init__(self, predict_fn=None, train_data=None, categorical_names=None, disc_perc=None):
        super().__init__()

class CounterfactualExplainer(_DummyAlibiExplainer):

    def __init__(
        self,
        predict_fn=None,
        feature_names=None,
        shape=None,
        drop_features=None,
        distance_fn=None,
        target_proba=None,
        target_class=None,
        max_iter=None,
        early_stop=None,
        lam_init=None,
        max_lam_steps=None,
        tol=None,
        learning_rate_init=None,
        feature_range=None,
        eps=None,
        init=None,
        decay=None,
        write_dir=None,
        debug=None,
        sess=None
    ):
        super().__init__()

class CounterfactualProtoExplainer(_DummyAlibiExplainer):

    def __init__(
        self,
        predict_fn=None,
        train_data=None,
        feature_names=None,
        shape=None,
        drop_features=None,
        kappa=None,
        beta=None,
        feature_range=None,
        gamma=None,
        ae_model=None,
        enc_model=None,
        theta=None,
        cat_vars=None,
        ohe=None,
        use_kdtree=None,
        learning_rate_init=None,
        max_iterations=None,
        c_init=None,
        c_steps=None,
        eps=None,
        clip=None,
        update_num_grad=None,
        write_dir=None,
        sess=None,
        trustscore_kwargs=None,
        d_type=None,
        w=None,
        disc_perc=None,
        standardize_cat_vars=None,
        smooth=None,
        center=None,
        update_feature_range=None
    ):
        super().__init__()



__version__ = '0.0.3'

from .explainers.counter_fact_basic import CounterFactualExplainerBasic
from .explainers.shuffle_importance import ShuffleImportanceExplainer
from .explainers.explainer import MercuryExplainer
from .explainers.partial_dependence import PartialDependenceExplainer
from .explanations.anchors import AnchorsWithImportanceExplanation
from .explanations.counter_factual import CounterfactualWithImportanceExplanation

from .explainers import ALEExplainer
from .explainers import AnchorsWithImportanceExplainer
from .explainers import CounterfactualExplainer, CounterfactualProtoExplainer


import numpy as np
import matplotlib.pyplot as plt


class FeatureImportanceExplanation():
    def __init__(self, data:dict, reverse:bool = False):
        """
        This class holds the data related to the importance a given
        feature has for a model.

        Args:
            data (dict):
                Contains the result of the PartialDependenceExplainer. It must be in the
                form of: ::
                    {
                        'feature_name': 1.0,
                        'feature_name2': 2.3, ...
                    }

            reverse (bool):
                Whether to reverse sort the features by increasing order (i.e. Worst
                performance (latest) = Smallest value). Default False (decreasing order).
        """
        self.data = data
        self._sorted_features = sorted(list(data.items()), key=lambda i: i[1],
                                       reverse=not reverse)

    def plot(self, ax: "matplotlib.axes.Axes" = None,  # noqa:F821
             figsize: tuple = (15, 15), limit_axis_x=False, **kwargs) -> "matplotlib.axes.Axes":  # noqa:F821
        """
        Plots a summary of the importances for each feature

        Args:
            figsize (tuple): Size of the plotted figure
            limit_axis_x (bool): Whether to adjust axis x to limit between the minimum and maximum feature values
        """
        ax = ax if ax else plt.gca()

        feature_names = [i[0] for i in self._sorted_features]
        feature_values = [i[1] for i in self._sorted_features]
        ax.barh(feature_names, feature_values)

        if limit_axis_x:
            ax.set_xlim(min(feature_values), max(feature_values))

        return ax

    def __getitem__(self, key:str)->float:
        """
        Gets the feature importance of the desired feature.

        Args:
            key (str): Name of the feature.
        """
        return self.data[key]

    def get_importances(self)->list:
        """ Returns a list of tuples (feature, importance) sorted by importances.
        """
        return self._sorted_features



"""
Optimization strategies for simple counterfactual explanations.
"""

import random
import typing as TP
from abc import ABCMeta, abstractmethod
import queue
import sys

from simanneal import Annealer
import numpy as np
import pandas as pd


class Strategy(metaclass=ABCMeta):
    """ Base class for explanation strategies. """

    def __init__(self,
                 state: 'np.ndarray',
                 bounds: 'np.ndarray',
                 step: 'np.ndarray',
                 fn: TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], 'np.ndarray'],
                 class_idx: int = 1,
                 threshold: float = 0.,
                 kernel: TP.Optional['np.ndarray'] = None,
                 report: bool = False) -> None:

        # Starting point
        self.state = state
        # Dimension bounds
        assert bounds.shape[0] == state.shape[0]
        self.bounds = bounds
        assert step.shape[0] == bounds.shape[0]
        self.step = step
        # Classifier fn
        self.fn = fn
        # Class
        self.class_idx = class_idx
        # Explored points
        self.explored = [self.state]
        # Probability threshold
        assert threshold >= 0. and threshold <= 1.
        self.threshold = threshold
        # Energies
        self.energies = [self.fn(self.state.reshape(1, -1))[:, self.class_idx][0]]
        # Kernel
        if kernel is None:
            kernel = np.ones(self.state.shape[0])
        # minimization or maximization problem?
        cur_prob = self.fn(self.state.reshape(1, -1))[0, self.class_idx]
        if cur_prob > self.threshold:
            self.min = True
        else:
            self.min = False

        self.kernel = kernel
        self.report = report

    @abstractmethod
    def run(self, *args, **kwargs) -> TP.Tuple['np.ndarray', float, 'np.ndarray', TP.Optional['np.ndarray']]:
        pass

    @abstractmethod
    def update(self, *args, **kwargs) -> None:
        pass


class SimulatedAnnealing(Strategy, Annealer):
    """
    Simulated Annealing strategy.

    Args:
        state (np.ndarray):
            Initial state (initial starting point).
        bounds (np.ndarray):
            Bounds to be used when moving around the probability space defined by `fn`.
        step (np.ndarray):
            Step size values to be used when moving around the probability space defined by `fn`.
            Lower values may take more time/steps to find a solution while too large values may
            make impossible to find a solution.
        fn (TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], 'np.ndarray']):
            Classifier `predict_proba`- like function.
        class_idx (int):
            Class to be explained (e.g. 1 for binary classifiers). Default value is 1.
        threshold (float):
            Probability to be achieved (if path is found). Default value is 0.0.
        kernel (TP.Optional['np.ndarray']):
            Used to penalize certain dimensions when trying to move around the probability
            space (some dimensions may be more difficult to explain, hence don't move along them).
        report (bool):
            Whether to display probability updates during space search.
    """

    def __init__(self,
                 state: 'np.ndarray',
                 bounds: 'np.ndarray',
                 step: 'np.ndarray',
                 fn: TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], 'np.ndarray'],
                 class_idx: int = 1,
                 threshold: float = 0.,
                 kernel: TP.Optional['np.ndarray'] = None,
                 report: bool = False) -> None:
        Strategy.__init__(self, state, bounds, step, fn, class_idx, threshold, kernel, report)
        # Annealer's __init__ uses signal, see Annealers's init:
        # https://github.com/perrygeo/simanneal/blob/b2576eb75d88f8b8c91d959a44dd708706bb108e/simanneal/anneal.py#L52
        # This may break execution (depending on where simanneal runs, e.g. threads)
        # Overload it!
        self.state = Annealer.copy_state(self, state)

    def move(self) -> None:
        """ Move step in Simulated Annealing. """
        self.state = np.random.uniform(-0.1, 0.1, size=self.state.shape[0]) * self.kernel * self.step + self.state
        for i, bound in enumerate(self.bounds):
            if self.kernel[i] != 0:
                if self.state[i] < bound[0]:
                    self.state[i] = bound[0]
                elif self.state[i] > bound[1]:
                    self.state[i] = bound[1]
        self.explored.append(self.state)

    def energy(self) -> None:
        """ Energy step in Simulated Annealing. """
        p = self.fn(np.array(self.state.reshape(1, -1)))[0, self.class_idx]
        # energy check
        if self.min:
            if self.threshold is not None and p < self.threshold:
                p = self.threshold
        else:
            if self.threshold is not None and p > self.threshold:
                p = self.threshold
        self.energies.append(p)
        value = p if self.min else -p
        return value

    def best_solution(self, n: int = 3) -> 'np.ndarray':
        """ Returns the n best solutions found during the Simulated Annealing.

        Args:
            n (int): Number of solutions to be retrieved. Default value is 3.
        """

        ps = self.fn(np.array(self.explored))[:, self.class_idx]
        sorted_ps_idxs = np.argsort(ps)
        if not self.min:
            sorted_ps_idxs = sorted_ps_idxs[::-1]
        return np.array(self.explored)[sorted_ps_idxs[:n]]

    def run(self, *args, **kwargs) -> TP.Tuple['np.ndarray', float, 'np.ndarray', TP.Optional['np.ndarray']]:
        """ Kick off Simulated Annealing.

        Args:
            **kargs:
                Simulated Annealing specific arguments
                    - tmin: min temperature
                    - tmax: max temperature
                    - steps: number of iterations

        Returns:
            result (TP.Tuple['np.ndarray', float, 'np.ndarray', TP.Optional['np.ndarray']]):
                Tuple containing found solution, probability achieved, points visited and
                corresponding energies (i.e. probabilities).
        """

        if 'tmin' in kwargs:
            self.Tmin = kwargs['tmin']
        if 'tmax' in kwargs:
            self.Tmax = kwargs['tmax']
        if 'steps' in kwargs:
            self.steps = kwargs['steps']
            self.updates = self.steps
            # self.updates = 100
        sol, p = self.anneal()  # type: TP.Tuple['np.ndarray', float]
        return sol, p, np.array(self.explored), np.array(self.energies)

    def update(self, *args, **kwargs) -> None:
        # TODO: By default use default's simanneal update method if necessary
        if self.report:
            Annealer.update(self, *args, **kwargs)


class MyPriorityQueue(queue.PriorityQueue):

    def __init__(self):
        super().__init__()

    def get_same_priority(self, limit=None, block=False, shuffle_limit=False):
        e = self.get(block=block)
        elements = [e]
        prio = e[0]
        while self.qsize():
            e = self.get(block=block)
            if e[0] != prio:
                self.put(e)
                break
            elements.append(e)
            # if we already reached the limit and no shuffle, we stop here
            if (limit is not None) and (not shuffle_limit) and (len(elements) >= limit):
                return elements
        if limit is not None and len(elements) > limit:
            random.shuffle(elements)
            for e in elements[limit:]:
                self.put(e)
            return elements[:limit]
        return elements


class Backtracking(Strategy):
    """ 
    Backtracking strategy.

    Args:
        state ('np.ndarray'): Initial state (initial starting point).
        bounds (np.ndarray): Bounds to be used when moving around the probability space defined by `fn`.
        step (np.ndarray): Step size values to be used when moving around the probability space defined by `fn`.
            Lower values may take more time/steps to find a solution while too large values may
            make impossible to find a solution.
        fn (TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], 'np.ndarray']): Classifier `predict_proba`- like function.
        class_idx (int): Class to be explained (e.g. 1 for binary classifiers). Default value is 1.
        threshold (float): Probability to be achieved (if path is found). Default value is 0.0.
        kernel (TP.Optional['np.ndarray']): Used to penalize certain dimensions when trying to move around the probability
            space (some dimensions may be more difficult to explain, hence don't move along them).
        report (bool): Whether to display probability updates during space search.
        keep_explored_points (bool): Whether to keep the points that the algorithm explores. Setting it to False will
            decrease the computation time and memory usage in some cases. Default value is True.

    """

    def __init__(self,
                 state: 'np.ndarray',
                 bounds: 'np.ndarray',
                 step: 'np.ndarray',
                 fn: TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], 'np.ndarray'],
                 class_idx: int = 1,
                 threshold: float = 0.,
                 kernel: TP.Optional['np.ndarray'] = None,
                 report: bool = False,
                 keep_explored_points: bool = True,
        ) -> None:
        super().__init__(state, bounds, step, fn, class_idx, threshold, kernel, report)
        self.keep_explored_points = keep_explored_points

    def run(self, *args, **kwargs) -> TP.Tuple['np.ndarray', float, 'np.ndarray', TP.Optional['np.ndarray']]:
        """ Kick off the backtracking.

        Args:
            **kwargs: Backtracking specific arguments
                        - max_iter: max number of iterations

        Returns:
            result (TP.Tuple['np.ndarray', float, 'np.ndarray', TP.Optional['np.ndarray']]):
                Tuple containing found solution, probability achieved, points visited and
                corresponding energies (i.e. probabilities).
        """

        # Backtracking specific arguments
        if 'max_iter' in kwargs:
            max_iter = kwargs['max_iter']
        else:
            max_iter = 0
        if 'limit' in kwargs:
            limit = kwargs['limit']
        else:
            limit = None
        if 'shuffle_limit' in kwargs:
            shuffle_limit = kwargs['shuffle_limit']
        else:
            shuffle_limit = False

        # Backtracking starting point.
        point = self.state.reshape(1, -1)
        curr_prob = self.fn(point)[0, self.class_idx]
        best_point = point.copy()
        best_prob = curr_prob

        # Define condition (>= or <=) from current point and given threshold.
        # and the priority modifier (ascending or descending order) using prio_mod.
        if self.threshold > curr_prob:
            condition = float.__ge__
            prio_mod = lambda x: -x
        else:
            condition = float.__le__
            prio_mod = lambda x: x
            self.kernel = 1.1 - self.kernel

        visited = np.hstack([point, [[curr_prob]]])
        explored = np.hstack([point, [[curr_prob]]]) if self.keep_explored_points else np.array([])

        q = MyPriorityQueue()  # type: ignore
        # Avoid duplicates and querying known points.
        cache = {str(point.tolist()): curr_prob}
        points = [point]
        counter = 0
        while condition(self.threshold, curr_prob) and (not max_iter or counter < max_iter):
            # Explore neighbours.
            neighbours = np.zeros((0, self.step.size))
            neighbours_l = []
            neighbours_kernel = []
            local_cache = set()
            for p_ in points:
                for i, delta in enumerate(self.step):
                    # In every direction.
                    for sign in (-1, 1):
                        aux = p_.copy()
                        new_aux_i = aux[0, i] + sign * delta
                        if new_aux_i >= self.bounds[i][0] and new_aux_i <= self.bounds[i][1]:
                            aux[0, i] = new_aux_i
                            aux_l = aux.tolist()
                            str_aux_l = str(aux_l)
                            if str_aux_l not in cache and str_aux_l not in local_cache:
                                # If point is not in cache (visited) -> enqueue.
                                neighbours = np.vstack([neighbours, aux])
                                neighbours_l.append(aux_l)
                                neighbours_kernel.append(self.kernel[i])
                                local_cache.add(str_aux_l)
            if len(neighbours_l):
                assert neighbours.shape[0] == len(neighbours_kernel) == len(neighbours_l), \
                    'Number of neighbours should match.'
                probs = self.fn(neighbours)[:, self.class_idx]
                for n_idx, kernel, pt in zip(
                        range(probs.shape[0]), neighbours_kernel, neighbours_l):
                    prob = float(probs[n_idx])
                    prio = prob * kernel
                    q.put((prio_mod(prio), prob, pt))
                    if self.keep_explored_points:
                        explored = np.vstack(
                            [explored,
                            np.hstack([neighbours[n_idx].reshape(1, -1), [[prob]]])])
                    cache[str(pt)] = prob
            try:
                elements = q.get_same_priority(limit=limit, block=False, shuffle_limit=shuffle_limit)
                curr_prob = max([x[1] for x in elements])
                points = [np.array(x[2]) for x in elements]
            except queue.Empty:
                self.visited = visited
                self.explored = explored
                return best_point[0], best_prob, self.visited, self.explored
            if condition(curr_prob, best_prob):
                best_prob = curr_prob
                best_point = points[0]
            for point in points:
                visited = np.vstack([visited, np.hstack([point,
                                                         [[curr_prob]]])])

            if self.report:

                iter_string = f"{counter}/{max_iter}"
                update_string = f"\r{iter_string}\t\t{round(best_prob, 2)}"
                if counter == 0:
                    print('\rIteration\tBest Prob', file=sys.stderr)
                    print(update_string, file=sys.stderr, end="")
                else:
                    print(update_string, file=sys.stderr, end="")
                sys.stderr.flush()

            counter += 1 if max_iter else 0

        self.explored = explored
        self.visited = visited
        return best_point[0], best_prob, self.visited, self.explored

    def update(self, *args, **kwargs):
        if self.report:
            # TODO: An implementation for updates should be provided
            pass



from abc import ABC, abstractmethod

import os
import dill

class MercuryExplainer(ABC):

    @abstractmethod
    def explain(self, data):
        pass

    def save(self, filename: str = "explainer.pkl"):
        """
        Saves the explainer with its internal state to a file.

        Args:
            filename (str): Path where the explainer will be saved
        """
        with open(filename, 'wb') as f:
            dill.dump(self, f)
        assert os.path.isfile(filename), "Error storing file"

    @classmethod
    def load(self, filename: str = "explainer.pkl"):
        """
        Loads a previosly saved explainer with its internal state to a file.

        Args:
            filename (str): Path where the explainer is stored
        """
        assert os.path.isfile(filename), "File does not exist or not a valid file"
        with open(filename, 'rb') as f:
            return dill.load(f)




import typing as TP
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import copy
import math

from typing import no_type_check
from itertools import count
from matplotlib.gridspec import GridSpec
from alibi.api.interfaces import Explanation, Explainer
from alibi.api.defaults import DEFAULT_META_ALE, DEFAULT_DATA_ALE
from alibi.explainers.ale import get_quantiles, bisect_fun, minimum_satisfied, adaptive_grid, ale_num

from .explainer import MercuryExplainer


class ALEExplainer(Explainer, MercuryExplainer):
    """
    Accumulated Local Effects for tabular datasets. Current implementation supports first order
    feature effects of numerical features.

    Args:
        predictor:
            A callable that takes in an NxF array as input and outputs an NxT array (N - number of
            data points, F - number of features, T - number of outputs/targets (e.g. 1 for single output
            regression, >=2 for classification).
        target_names:
            A list of target/output names used for displaying results.
    """

    def __init__(self, predictor: TP.Callable, target_names: TP.Union[TP.List[str], str]) -> None:
        super().__init__(meta=copy.deepcopy(DEFAULT_META_ALE))

        if (not isinstance(target_names, list) and
                not isinstance(target_names, str)):
            raise AttributeError('The attribute target_names should be a string or list.')

        if type(target_names) == str:
            target_names = [target_names]

        self.predictor = predictor
        self.target_names = target_names

    def explain(self, X: pd.DataFrame, min_bin_points: int = 4, ignore_features: list = []) -> Explanation:
        """
        Calculate the ALE curves for each feature with respect to the dataset `X`.

        Args:
            X:
                An NxF tabular dataset used to calculate the ALE curves. This is typically the training dataset
                or a representative sample.
            min_bin_points:
                Minimum number of points each discretized interval should contain to ensure more precise
                ALE estimation.
            ignore_features:
                Features that will be ignored while computing the ALE curves. Useful for reducing computing time
                if there are predictors we dont care about.

        Returns:
            An `Explanation` object containing the data and the metadata of the calculated ALE curves.

        """
        self.meta['params'].update(min_bin_points=min_bin_points)

        if not isinstance(X, pd.DataFrame):
            raise ValueError('X must be a pandas DataFrame')

        features = list(X.columns)
        n_features =len(features)

        self.feature_names = np.array(features)
        self.target_names = np.array(self.target_names)

        feature_values = []
        ale_values = []
        ale0 = []
        feature_deciles = []

        X = X[features].values

        # TODO: use joblib to paralelise?
        for feature, feat_name in enumerate(self.feature_names):
            if feat_name not in ignore_features:
                q, ale, a0 = ale_num(
                    self.predictor,
                    X=X,
                    feature=feature,
                    min_bin_points=min_bin_points
                )
                deciles = get_quantiles(X[:, feature], num_quantiles=11)

                feature_values.append(q)
                ale_values.append(ale)
                ale0.append(a0)
                feature_deciles.append(deciles)

        constant_value = self.predictor(X).mean()
        # TODO: an ALE plot ideally requires a rugplot to gauge density of instances in the feature space.
        # I've replaced this with feature deciles which is coarser but has constant space complexity
        # as opposed to a rugplot. Alternatively, could consider subsampling to produce a rug with some
        # maximum number of points.
        return self._build_explanation(
            ale_values=ale_values,
            ale0=ale0,
            constant_value=constant_value,
            feature_values=feature_values,
            feature_deciles=feature_deciles,
            ignore_features=ignore_features
        )

    def _build_explanation(self,
                          ale_values: TP.List[np.ndarray],
                          ale0: TP.List[np.ndarray],
                          constant_value: float,
                          feature_values: TP.List[np.ndarray],
                          feature_deciles: TP.List[np.ndarray],
                          ignore_features: TP.List = []) -> Explanation:
        """
        Helper method to build the Explanation object.
        """
        # TODO decide on the format for these lists of arrays
        # Currently each list element relates to a feature and each column relates to an output dimension,
        # this is different from e.g. SHAP but arguably more convenient for ALE.

        data = copy.deepcopy(DEFAULT_DATA_ALE)
        data.update(
            ale_values=ale_values,
            ale0=ale0,
            constant_value=constant_value,
            feature_values=feature_values,
            feature_names=[x for x in self.feature_names if x not in ignore_features],
            target_names=self.target_names,
            feature_deciles=feature_deciles
        )

        return Explanation(meta=copy.deepcopy(self.meta), data=data)
    
    def save(self, filename):
        """Overwrite to ensure that we use MercuryExplainer.save"""
        MercuryExplainer.save(self, filename=filename)

@no_type_check
def plot_ale(exp: Explanation,
             features: TP.Union[TP.List[TP.Union[int, str]], str] = 'all',
             targets: TP.Union[TP.List[TP.Union[int, str]], str] = 'all',
             n_cols: int = 3,
             sharey: str = 'all',
             constant: bool = False,
             ax: TP.Union['plt.Axes', np.ndarray, None] = None,
             line_kw: TP.Optional[dict] = None,
             fig_kw: TP.Optional[dict] = None) -> 'np.ndarray':
    """
    Plot ALE curves on matplotlib axes.

    Args:
        exp: An `Explanation` object produced by a call to the `ALE.explain` method.
        features: A list of features for which to plot the ALE curves or `all` for all features.
            Can be a mix of integers denoting feature index or strings denoting entries in
            `exp.feature_names`. Defaults to 'all'.
        targets: A list of targets for which to plot the ALE curves or `all` for all targets.
            Can be a mix of integers denoting target index or strings denoting entries in
            `exp.target_names`. Defaults to 'all'.
        n_cols: Number of columns to organize the resulting plot into.
        sharey: A parameter specifying whether the y-axis of the ALE curves should be on the same scale
            for several features. Possible values are `all`, `row`, `None`.
        constant: A parameter specifying whether the constant zeroth order effects should be added to the
            ALE first order effects.
        ax: A `matplotlib` axes object or a numpy array of `matplotlib` axes to plot on.
        line_kw: Keyword arguments passed to the `plt.plot` function.
        fig_kw: Keyword arguments passed to the `fig.set` function.

    Returns:
        An array of matplotlib axes with the resulting ALE plots.

    """
    # line_kw and fig_kw values
    default_line_kw = {'markersize': 3, 'marker': 'o', 'label': None}
    if line_kw is None:
        line_kw = {}
    line_kw = {**default_line_kw, **line_kw}

    default_fig_kw = {'tight_layout': 'tight'}
    if fig_kw is None:
        fig_kw = {}
    fig_kw = {**default_fig_kw, **fig_kw}

    if features == 'all':
        selected_features = range(0, len(exp.feature_names))
    else:
        selected_features = []
        for ix, f in enumerate(features):
            if isinstance(f, str):
                try:
                    selected_features.append(exp.feature_names.index(f))
                except ValueError:
                    raise ValueError("Feature name {} does not exist.".format(f))
            elif isinstance(f, int):
                selected_features.append(f)
    n_features = len(selected_features)

    if targets == 'all':
        targets = range(0, len(exp.target_names))
    else:
        for ix, t in enumerate(targets):
            if isinstance(t, str):
                try:
                    t = np.argwhere(exp.target_names == t).item()
                except ValueError:
                    raise ValueError("Target name {} does not exist.".format(t))
            targets[ix] = t

    # make axes
    if ax is None:
        fig, ax = plt.subplots()

    if isinstance(ax, plt.Axes) and n_features != 1:
        ax.set_axis_off()  # treat passed axis as a canvas for subplots
        fig = ax.figure
        n_cols = min(n_cols, n_features)
        n_rows = math.ceil(n_features / n_cols)

        axes = np.empty((n_rows, n_cols), dtype=np.object_)
        axes_ravel = axes.ravel()
        # gs = GridSpecFromSubplotSpec(n_rows, n_cols, subplot_spec=ax.get_subplotspec())
        gs = GridSpec(n_rows, n_cols)
        for i, spec in zip(range(n_features), gs):
            # determine which y-axes should be shared
            if sharey == 'all':
                cond = i != 0
            elif sharey == 'row':
                cond = i % n_cols != 0
            else:
                cond = False

            if cond:
                axes_ravel[i] = fig.add_subplot(spec, sharey=axes_ravel[i - 1])
                continue
            axes_ravel[i] = fig.add_subplot(spec)

    else:  # array-like
        if isinstance(ax, plt.Axes):
            ax = np.array(ax)
        if ax.size < n_features:
            raise ValueError("Expected ax to have {} axes, got {}".format(n_features, ax.size))
        axes = np.atleast_2d(ax)
        axes_ravel = axes.ravel()
        fig = axes_ravel[0].figure

    # make plots
    for ix, feature, ax_ravel in \
            zip(count(), selected_features, axes_ravel):
        _ = _plot_one_ale_num(exp=exp,
                              feature=feature,
                              targets=targets,
                              constant=constant,
                              ax=ax_ravel,
                              legend=not ix,  # only one legend
                              line_kw=line_kw)

    # if explicit labels passed, handle the legend here as the axis passed might be repeated
    if line_kw['label'] is not None:
        axes_ravel[0].legend()

    fig.set(**fig_kw)
    # TODO: should we return just axes or ax + axes
    return axes

@no_type_check
def _plot_one_ale_num(exp: Explanation,
                      feature: int,
                      targets: TP.List[int],
                      constant: bool = False,
                      ax: 'plt.Axes' = None,
                      legend: bool = True,
                      line_kw: dict = None) -> 'plt.Axes':
    """
    Plots the ALE of exactly one feature on one axes.
    """
    import matplotlib.pyplot as plt
    from matplotlib import transforms

    if ax is None:
        ax = plt.gca()

    # add zero baseline
    ax.axhline(0, color='grey')

    # Sometimes within the computation of the ale values, we get more values corresponding
    # to the feature than values corresponding to the target or vice-versa, i.e. the number
    # of X's and Y's is not the same and therefore is not possible to properly build the ale
    # plot. These conditions help ensure len(x) equals len(y).
    if len(exp.feature_values[feature]) == len(exp.ale_values[feature][:, targets]):
        lines = ax.plot(
            exp.feature_values[feature],
            exp.ale_values[feature][:, targets] + constant * exp.constant_value,
            **line_kw
        )
    elif len(exp.feature_values[feature]) < len(exp.ale_values[feature][:, targets]):
        diff = len(exp.ale_values[feature][:, targets]) - len(exp.feature_values[feature])
        x = np.append(exp.feature_values[feature], np.repeat(exp.feature_values[feature][-1], diff))
        y = exp.ale_values[feature][:, targets]
        lines = ax.plot(
            x, y, **line_kw
        )
    elif len(exp.feature_values[feature]) > len(exp.ale_values[feature][:, targets]):
        diff = len(exp.feature_values[feature]) - len(exp.ale_values[feature][:, targets])
        y = np.append(exp.ale_values[feature][:, targets], np.repeat(exp.ale_values[feature][:, targets][-1], diff))
        x = exp.feature_values[feature]
        lines = ax.plot(
            x, y, **line_kw
        )

    # add decile markers to the bottom of the plot
    trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
    ax.vlines(exp.feature_deciles[feature][1:], 0, 0.05, transform=trans)

    ax.set_xlabel(exp.feature_names[feature])
    ax.set_ylabel('ALE')

    if legend:
        # if no explicit labels passed, just use target names
        if line_kw['label'] is None:
            ax.legend(lines, exp.target_names[targets])

    return ax



import typing as TP
import numpy as np
import pandas as pd
import tensorflow as tf
tf.compat.v1.disable_eager_execution()

from abc import ABC
from .explainer import MercuryExplainer

from mercury.explainability.explanations.counter_factual import (
    CounterfactualWithImportanceExplanation
)

from alibi.api.interfaces import Explanation
from alibi.explainers import CounterFactual, CounterFactualProto


class CounterFactualExplainerBase(ABC):
    """
    Abstract base class for CounterFactual explainers which hold common
    functionalities. You should extend this if you are implementing a
    custom counter factual explainer.

    Args:
        feature_names:
            List of names of the features used in the model.
        drop_features:
            List with the names, as specified in feature_names, or with the
            indexes of the variables to leave out while looking for
            explanations.
    """

    def __init__(self, feature_names: TP.List[str],
                 drop_features: TP.List[TP.Any] = []):
        self.feature_names = feature_names
        self.drop_features = drop_features
        # Each subclass will use its internal Alibi member (CounterFactual or
        # CounterFactualProto)
        self.cf_super = None

    def _regularize_counterfactual(
        self,
        predict_fn: TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], TP.Union[float, 'np.ndarray']],
        explain_data: 'np.ndarray',
        counterfactual_original: 'np.ndarray',
        min_prob: float,
        threshold_incr: float = 0.01,
        init_threshold: float = 0.1,
        max_iterations: int = 10000) -> TP.Tuple[TP.Any]:
        """
        Internal function to regularize the encountered counterfactuals.

        Args:
            predict_fn:
                Model prediction function. This function should return the probabilities of belonging to each class.
            explain_data:
                Numpy array with containing the values for all features for the original instance.
            counterfactual_original:
                Numpy array containing the resulting counterfactual computed by alibi.
            min_prob:
                Target minimum probability to be reached while regularizing.
            threshold_incr:
                Increments to be applied to the threshold by which each feature is reduced to zero. Defaults to 0.01.
            init_threshold:
                Initial threshold that determines whether the feature has to be reduced to zero or not. Defaults 0.1.
            max_iterations:
                Maximum number of iterations for which the regularization process runs. Defaults to 10000.

            Returns:
                A tuple containing the new counterfactual, the difference between the new counterfactual and the original
                instance and finally, the achieved probability.
        """
        try:
            if len(explain_data[0]) != len(counterfactual_original[0]):
                raise AttributeError("""
                The dimensions of the original instance ({}) and that
                of the counterfactual instance ({}) do not match
                """.format(len(explain_data[0]), len(counterfactual_original[0])))
        except AttributeError:
            raise

        differences = (explain_data - counterfactual_original)[0]
        prediction = predict_fn(counterfactual_original)[0]
        counter_class = np.argmax(prediction)
        counter_proba = prediction[counter_class]

        try:
            if min_prob >= counter_proba:
                raise AttributeError("""
                The objective class probability is larger than that of the counterfactual passed to this method.
                This means that either the target probability is greater than that stablished when calling the
                counterfactual method or that the counterfactual being passed to this method is not well formed.
                The inputed target_probability was {:.3f} while the obtained probability for this counterfactual is {:.3f}
                """.format(min_prob, counter_proba))
        except AttributeError:
            raise

        new_proba = counter_proba
        new_cfs = []
        threshold = init_threshold
        n_iter = 0

        while new_proba >= min_prob:
            new_cf = []
            for i, diff in enumerate(differences):
                if np.abs(diff) > threshold:
                    new_cf.append(counterfactual_original[0][i])
                else:
                    new_cf.append(explain_data[0][i])
            new_probas_ = predict_fn(np.array(new_cf).reshape(1, -1))[0]
            new_proba_ = new_probas_[counter_class]
            if new_proba_ > min_prob:
                new_cfs.append(new_cf)
            else:
                if len(new_cfs) == 0:
                    new_cfs.append(counterfactual_original[0])

            differences = (explain_data - new_cfs[-1])[0]
            threshold += threshold_incr
            new_proba = new_proba_
            n_iter += 1

            if n_iter == max_iterations:
                print('No solution...')
                print('Final threshold: ', threshold)
                break

        regularized_counterfactual = np.array(new_cfs[-1])
        diff = explain_data - regularized_counterfactual
        new_probability = predict_fn(regularized_counterfactual.reshape(1, -1))[0][counter_class]

        return regularized_counterfactual, diff, new_probability

    def _compute_feature_range(
        self,
        feature_range: TP.Union[TP.Tuple, str],
        drop_features: TP.List[int],
        explain_data: 'np.ndarray'
    ) -> 'np.ndarray':
        """
        This function computes a feature range adjusted to a given observation
        such that the range for the features established on argument drop_features
        do not allow for change.

        Args:
            feature_range:
                Tuple with min and max ranges to allow for perturbed instances. Min and max
                ranges have to be numpy arrays with dimension (1 x nb of features)
                for feature-wise ranges when dropping features.
            drop_features:
                List with the indexes of the variables to leave out while looking
                for explanations.
            explain_data:
                Instance to be explained

        Returns:
            Tuple containing the new feature range.
        """
        explain_data = explain_data.reshape(1, -1)
        feat_range = feature_range
        feat_range[0][drop_features] = explain_data[0][drop_features]
        feat_range[1][drop_features] = explain_data[0][drop_features]
        return feat_range

    def get_feature_importance(
        self,
        explain_data: pd.DataFrame,
        n_important_features: int = 3,
        print_every: int = 5,
        threshold_probability=0.5,
        threshold_incr: float = 0.01,
        init_threshold: float = 0.1,
        max_iterations: int = 10000,
        get_report: bool = True
    ) -> CounterfactualWithImportanceExplanation:
        """
        This method computes the feature importance for a set of observations.

        Args:
            explain_data:
                Pandas DataFrame containing all the instances for which to find an anchor and therefore
                obtain feature importances.
            n_important_features:
                Number of top features that will be printed. Defaults to 3.
            print_every:
                Logging information (default=5).
            threshold_incr:
                The increment of the threshold to be used in the regularizer so as to bring the differences
                between a counterfactual instance and the original one to zero.
            init_threshold:
                The initial and minimum value to be used in the regularizer to bring the differences
                between a counterfactual instance and the original one to zero.
            max_iterations:
                Maximum number of iterations for the regularizer. This parameter gets pass
                down to the regularize_counterfactual method.
            get_report:
                Boolean determining whether to print the explanation or not.

        Returns:
            A CounterfactualExtendedExplanation object that contains all explanations as well as
            their interpretation.
        """
        failed_explanations = 0
        counterfactuals = []
        attempted_explanations = 0
        for item_idx, item in explain_data.iterrows():
            try:
                attempted_explanations += 1
                cf_explanation = self.explain(
                    item.values.reshape(1, -1)
                )
                cf_regularized, diff, new_proba = self._regularize_counterfactual(
                    self.predict_fn,
                    item.values.reshape(1, -1),
                    cf_explanation.data['cf']['X'],
                    min_prob=threshold_probability,
                    threshold_incr=threshold_incr,
                    max_iterations=max_iterations,
                    init_threshold=init_threshold
                )
                cf_explanation = {
                    'cf_original': cf_explanation.data['cf']['X'],
                    'cf_class': cf_explanation.data['cf']['class'],
                    'cf_proba': cf_explanation.data['cf']['proba'][0][
                        cf_explanation.data['cf']['class']
                    ],
                    'cf_regularized': cf_regularized,
                    'cf_reg_proba': new_proba,
                    'diff_orig_reg': diff
                }
            except Exception:
                failed_explanations += 1
                print('''
                There's been a problem while computing the explanation for observation: {}.
                If this observation is critical to you try with some other hyperparameters.
                '''.format(item_idx))
                cf_explanation = {
                    'cf_original': None,
                    'cf_class': None,
                    'cf_proba': None,
                    'cf_regularized': None,
                    'cf_reg_proba': None,
                    'diff_orig_reg': None
                }
            if attempted_explanations % print_every == 0:
                print('{} counterfactuals already computed. Found a solution for {} of them'.format(
                    attempted_explanations, attempted_explanations - failed_explanations
                ))
            counterfactuals.append(cf_explanation)
        print('All counterfactuals have been computed and regularized')

        count_differences = dict(
            (feat_name, 0.0) for feat_name in self.feature_names
        )
        total_diffs = np.zeros(len(self.feature_names))

        if len(counterfactuals) > 0:
            for counterfactual_i in counterfactuals:
                if counterfactual_i['cf_original'] is not None:
                    total_diffs += counterfactual_i['diff_orig_reg'][0]
                    diffs_dict = dict(zip(self.feature_names, counterfactual_i['diff_orig_reg'][0]))
                    for k, v in diffs_dict.items():
                        if np.abs(v) > 0.0:
                            count_differences[k] += 1

        max_diff = np.abs(total_diffs).max(axis=0)
        total_diffs /= max_diff

        importances = [
            (name, value, direction)
            for value, direction, name
            in sorted(
                zip(np.abs(total_diffs), np.sign(total_diffs), self.feature_names),
                reverse=True
            )
        ]

        count_differences_norm = dict(
            (key, value / (explain_data.shape[0] - failed_explanations))
            for key, value
            in count_differences.items()
        )
        count_differences_norm = {
            key: value
            for key, value
            in sorted(
                count_differences_norm.items(),
                key=lambda item: item[1],
                reverse=True
            )
        }

        n_important_features = n_important_features if n_important_features < len(self.feature_names) else len(self.feature_names)

        counterfactualExtendedExplanation = CounterfactualWithImportanceExplanation(
            explain_data,
            counterfactuals,
            importances,
            count_differences,
            count_differences_norm
        )

        if get_report:
            if n_important_features == len(self.feature_names):
                print('The total number of important features was too large and therefore all will be shown')
            if failed_explanations > 0:
                print('There were a total of {:d} fails'.format(failed_explanations))
            counterfactualExtendedExplanation.interpret_explanations(
                n_important_features
            )
        return counterfactualExtendedExplanation

    def __new__(cls, *args, **kwargs):
        if cls is CounterFactualExplainerBase:
            raise TypeError("""You shouldn't be instantiating this. """ +
                            """Use the proper subclass""")
        return object.__new__(cls)


class CounterfactualExplainer(CounterFactualExplainerBase):
    """
    Backed by Alibi's CounterFactual, this class extends its functionality to allow for the computation of Feature
    Importances by means of the computation of several Counterfactuals.

    Args:
        predict_fn:
            Model prediction function. This function should return the probabilities of belonging to each class.
        feature_names:
            List of names of the features used in the model.
        shape:
            Shape of input data starting with batch size. By default it is inferred from feature_names.
        drop_features:
            List with the names, as specified in feature_names, or with the indexes of the variables to leave
            out while looking for explanations (in get_feature_importance method).
        distance_fn:
            Distance function to use in the loss term
        target_proba:
            Target probability for the counterfactual to reach
        target_class:
            Target class for the counterfactual to reach, one of 'other', 'same' or an integer denoting
            desired class membership for the counterfactual instance
        max_iter:
            Maximum number of interations to run the gradient descent for (inner loop)
        early_stop:
            Number of steps after which to terminate gradient descent if all or none of found instances are solutions
        lam_init:
            Initial regularization constant for the prediction part of the Wachter loss
        max_lam_steps:
            Maximum number of times to adjust the regularization constant (outer loop) before terminating the search
        tol:
            Tolerance for the counterfactual target probability
        learning_rate_init:
            Initial learning rate for each outer loop of lambda
        feature_range:
            Tuple with min and max ranges to allow for perturbed instances. Min and max ranges can be floats or
            numpy arrays with dimension (1 x nb of features) for feature-wise ranges
        eps:
            Gradient step sizes used in calculating numerical gradients, defaults to a single value for all
            features, but can be passed an array for feature-wise step sizes
        init:
            Initialization method for the search of counterfactuals, currently must be 'identity'
        decay:
            Flag to decay learning rate to zero for each outer loop over lambda
        write_dir:
            Directory to write Tensorboard files to
        debug:
            Flag to write Tensorboard summaries for debugging
        sess:
            Optional Tensorflow session that will be used if passed instead of creating or inferring one internally
    """
    def __init__(
        self,
        predict_fn: TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], TP.Union[float, 'np.ndarray']],
        feature_names: TP.List[str],
        shape: TP.Tuple[TP.Any] = None,
        drop_features: TP.List[TP.Any] = [],
        distance_fn: str = 'l1',
        target_proba: float = 1.0,
        target_class: TP.Union[str, int] = 'other',
        max_iter: int = 1000,
        early_stop: int = 50,
        lam_init: float = 1e-1,
        max_lam_steps: int = 10,
        tol: float = 0.05,
        learning_rate_init: float = 0.1,
        feature_range: TP.Union[TP.Tuple[TP.Any], str] = (-1e10, 1e10),
        eps: TP.Union[float, 'np.ndarray'] = 0.01,
        init: str = 'identity',
        decay: bool = True,
        write_dir: str = None,
        debug: bool = False,
        sess: tf.compat.v1.Session = None
    ) -> None:
        super(CounterfactualExplainer,
              self).__init__(feature_names, drop_features)

        self.predict_fn = predict_fn
        self.shape = shape if shape else (1, len(feature_names))
        self.distance_fn = distance_fn
        self.target_proba = target_proba
        self.target_class = target_class
        self.max_iter = max_iter
        self.early_stop = early_stop
        self.lam_init = lam_init
        self.max_lam_steps = max_lam_steps
        self.tol = tol
        self.learning_rate_init = learning_rate_init
        self.feature_range = feature_range
        self.eps = eps
        self.init = init
        self.decay = decay
        self.write_dir = write_dir
        self.debug = debug
        self.sess = sess
        self.cat_vars = None

        # Instantiate counterfactual
        self.cf_super = CounterFactual(
            predict_fn=self.predict_fn, shape=self.shape,
            distance_fn=self.distance_fn, target_proba=self.target_proba,
            target_class=self.target_class, max_iter=self.max_iter,
            early_stop=self.early_stop, lam_init=self.lam_init,
            max_lam_steps=self.max_lam_steps, tol=self.tol,
            learning_rate_init=self.learning_rate_init,
            feature_range=self.feature_range, eps=self.eps,
            init=self.init, decay=self.decay, write_dir=self.write_dir,
            debug=self.debug, sess=self.sess)

    def explain(self, explain_data: 'np.ndarray') -> Explanation:
        """
        This method serves two purposes. Firstly as an interface between the user and the CounterFactual class
        explain method and secondly, if there are features that shouldn't be taken into account, the method
        recomputes feature_range to account for this and reinstantiates the parent class (CounterFactual).
        Explain an instance and return the counterfactual with metadata.

        Args:
            explain_data (np.ndarray): Instance to be explained

        Returns:
            Explanation object as specified in Alibi's original explain method.
        """
        return self.cf_super.explain(
            explain_data
        )


class CounterfactualProtoExplainer(CounterFactualExplainerBase):
    """
    Backed by Alibi's CounterFactualProto, this class extends its functionality to allow for the computation of Feature
    Importances by means of the computation of several Counterfactuals.

    Args:
        predict_fn:
            Model prediction function. This function should return the probabilities of belonging to
            each class.
        train_data:
            Numpy array or Pandas dataframe with the training examples
        feature_names:
            List of names of the features used in the model. Will be inferred by default from train_data
        shape:
            Shape of input data starting with batch size. Will be inferred by default from train_data
        drop_features:
            List with the names, as specified in feature_names, or with the indexes of the variables to leave
            out while looking for explanations.
        beta:
            Regularization constant for L1 loss term
        kappa:
            Confidence parameter for the attack loss term
        feature_range:
            Tuple with min and max ranges to allow for perturbed instances. Min and max ranges can be floats or
            numpy arrays with dimension (1x nb of features) for feature-wise ranges.
            Will be inferred by default from train_data
        gamma:
            Regularization constant for optional auto-encoder loss term
        ae_model:
            Optional auto-encoder model used for loss regularization
        enc_model:
            Optional encoder model used to guide instance perturbations towards a class prototype
        theta:
            Constant for the prototype search loss term. Default is 5. Set it to zero to disable it.
        cat_vars:
            Dict with as keys the categorical columns and as values
            the number of categories per categorical variable.
        ohe:
            Whether the categorical variables are one-hot encoded (OHE) or not. If not OHE, they are
            assumed to have ordinal encodings.
        use_kdtree:
            Whether to use k-d trees for the prototype loss term if no encoder is available
        learning_rate_init:
            Initial learning rate of optimizer
        max_iterations:
            Maximum number of iterations for finding a counterfactual
        c_init:
            Initial value to scale the attack loss term. If the computation shall be fastened up, this
            parameter should take a small value (0 or 1).
        c_steps:
            Number of iterations to adjust the constant scaling the attack loss term. If the computation
            shall be fastened up this parameter should take a somewhat small value (between 1 and 5).
        eps:
            If numerical gradients are used to compute dL/dx = (dL/dp) * (dp/dx), then eps[0] is used to
            calculate dL/dp and eps[1] is used for dp/dx. eps[0] and eps[1] can be a combination of float values and
            numpy arrays. For eps[0], the array dimension should be (1x nb of prediction categories) and for
            eps[1] it should be (1x nb of features)
        clip:
            Tuple with min and max clip ranges for both the numerical gradients and the gradients
            obtained from the TensorFlow graph
        update_num_grad:
            If numerical gradients are used, they will be updated every update_num_grad iterations. If
            the computation shall be fastened up this parameter should take a somewhat large value
            (between 100 and 250).
        update_num_grad:
            If numerical gradients are used, they will be updated every update_num_grad iterations
        write_dir:
            Directory to write tensorboard files to
        sess:
            Optional Tensorflow session that will be used if passed instead of creating or inferring one internally
        trustscore_kwargs:
            keyword arguments for the trust score object used to define the k-d trees for each class.
            (See original alibi) counterfactual guided by prototypes docs.
        d_type:
            Distance metric. Options are "abdm", "mvdm" or "abdm-mvdm".
        w:
            If the combined metric "abdm-mvdm" is used, w is the weight (between 0 and 1) given to abdm.
        disc_perc:
            List with percentiles (int) used for discretization
        standardize_cat_vars:
            Whether to return the standardized values for the numerical distances of each categorical feature.
        smooth:
            if the difference in the distances between the categorical variables is too large, then a lower
            value of the smooth argument (0, 1) can smoothen out this difference. This would only be relevant
            if one categorical variable has significantly larger differences between its categories than others.
            As a result, the counterfactual search process will likely leave that categorical variable unchanged.
        center:
            Whether to center the numerical distances of the categorical variables between the min and max feature ranges.
        update_feature_range:
            whether to update the feature_range parameter for the categorical variables based on the min and max values
            it computed in the constructor.
    """

    def __init__(self,
        predict_fn: TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], TP.Union[float, 'np.ndarray']],
        train_data: TP.Union[np.ndarray, pd.DataFrame],
        feature_names: TP.List[str] = None,
        shape: TP.Tuple[TP.Any] = None,
        drop_features: TP.List[TP.Any] = [],
        kappa: float = 0.,
        beta: float = .1,
        feature_range: TP.Union[TP.Tuple[TP.Any], str] = None,
        gamma: float = 0.,
        ae_model: tf.keras.Model = None,
        enc_model: tf.keras.Model = None,
        theta: float = 5.0,
        cat_vars: dict = None,
        ohe: bool = False,
        use_kdtree: bool = False,
        learning_rate_init: float = 1e-2,
        max_iterations: int = 1000,
        c_init: float = 10.,
        c_steps: int = 10,
        eps: TP.Union[float,'np.ndarray'] = (1e-3, 1e-3),
        clip: TP.Tuple[float] = (-1000, 1000),
        update_num_grad: int = 1,
        write_dir: str = None,
        sess: tf.compat.v1.Session = None,
        trustscore_kwargs: dict = None,
        d_type: str = 'abdm',
        w: float = None,
        disc_perc: TP.Union[TP.List[TP.Union[int, float]], TP.Tuple[TP.Union[int, float]]] = [25, 50, 75],
        standardize_cat_vars: bool = False,
        smooth: float = 1.,
        center: bool = True,
        update_feature_range: bool = True
    ) -> None:
        if feature_names is None and type(train_data) is not pd.DataFrame:
            raise ValueError("If feature_names is not present, train_data" +
                             "should be a Pandas DataFrame. ")

        super(CounterfactualProtoExplainer,
              self).__init__(feature_names, drop_features)

        self.fitted = False
        self.feature_range = feature_range
        self.predict_fn = predict_fn
        self.shape = (1, train_data.shape[-1])
        self.feature_names = feature_names
        self.kappa = kappa
        self.beta = beta
        self.gamma = gamma
        self.ae_model = ae_model
        self.enc_model = enc_model
        self.theta = theta
        self.cat_vars = cat_vars
        self.ohe = ohe
        self.use_kdtree = True if not enc_model and not ae_model else False
        self.learning_rate_init = learning_rate_init
        self.max_iterations = max_iterations
        self.c_init = c_init
        self.c_steps = c_steps
        self.eps = eps
        self.clip = clip
        self.update_num_grad = update_num_grad
        self.write_dir = write_dir
        self.sess = sess

        self.train_data = train_data

        if isinstance(train_data, pd.DataFrame):
            self.feature_names = list(train_data.columns)
            self.feature_range = \
                (train_data.min().values, train_data.max().values) \
                    if not feature_range else (-2e5, 2e5)
            self.train_data=train_data.values

        self.cf_super = CounterFactualProto(
            predict=predict_fn, shape=self.shape, kappa=self.kappa, beta=self.beta,
            feature_range=self.feature_range, gamma=self.gamma, ae_model=self.ae_model,
            enc_model=self.enc_model, theta=self.theta, cat_vars=self.cat_vars, ohe=self.ohe,
            use_kdtree=self.use_kdtree, learning_rate_init=self.learning_rate_init,
            max_iterations=self.max_iterations, c_init=self.c_init, c_steps=self.c_steps,
            eps=self.eps, clip=self.clip, update_num_grad=self.update_num_grad,
            write_dir=self.write_dir, sess=self.sess
        )

        self.cf_super = self.cf_super.fit(self.train_data, trustscore_kwargs=trustscore_kwargs, d_type=d_type,
            w=w, disc_perc=disc_perc, standardize_cat_vars=standardize_cat_vars,
            smooth=smooth, center=center, update_feature_range=update_feature_range)

    def explain(
        self,
        explain_data: 'np.ndarray',
        Y: 'np.ndarray' = None,
        target_class: list = None,
        k: int = None,
        k_type: str = 'mean',
        threshold: float = 0.,
        verbose: bool = False,
        print_every: int = 100,
        log_every: int = 100) -> Explanation:
        """
        This method serves three purposes. Firstly as an interface between the user and the CounterFactualProto
        class explain method, secondly, if there are features that shouldn't be taken into account, the method
        recomputes feature_range to account for this and reinstantiates the parent class (CounterFactualProto)
        and thridly, if a new CounterFactualProto instance has been created it's fitted.
        Explain instance and return counterfactual with metadata.

        Args:
            X:
                Instances to explain
            Y:
                Labels for X as one-hot-encoding
            target_class:
                List with target classes used to find closest prototype. If None, the nearest prototype
                except for the predict class on the instance is used.
            k:
                Number of nearest instances used to define the prototype for a class. Defaults to using all
                instances belonging to the class if an encoder is used and to 1 for k-d trees.
            k_type:
                Use either the average encoding of the k nearest instances in a class (k_type='mean') or
                the k-nearest encoding in the class (k_type='point') to define the prototype of that class.
                Only relevant if an encoder is used to define the prototypes.
            threshold:
                Threshold level for the ratio between the distance of the counterfactual to the prototype of the
                predicted class for the original instance over the distance to the prototype of the predicted class
                for the counterfactual. If the trust score is below the threshold, the proposed counterfactual does
                not meet the requirements.
            verbose:
                Print intermediate results of optimization if True
            print_every:
                Print frequency if verbose is True
            log_every:
                Tensorboard log frequency if write directory is specified

        Returns:
            explanation, Dictionary containing the counterfactual with additional metadata
        """

        return self.cf_super.explain(
            explain_data,
            Y,
            target_class,
            k,
            k_type,
            threshold,
            verbose,
            print_every,
            log_every
        )



import typing as TP
import numpy as np
import pandas as pd
import signal

from .explainer import MercuryExplainer
from mercury.explainability.explainers import run_until_timeout
from alibi.explainers import AnchorTabular
from mercury.explainability.explanations.anchors import AnchorsWithImportanceExplanation
from alibi.api.interfaces import Explanation


class AnchorsWithImportanceExplainer(AnchorTabular, MercuryExplainer):
    """
    Extending Alibi's AnchorsTabular Implementation, this module allows for the
    computation of feature importance by means of calculating several anchors.
    Initialize the anchor tabular explainer.

    Args:
        predict_fn: Model prediction function
        train_data: Pandas Dataframe with the features
        disc_perc: List or tuple with percentiles (int) used for discretization.
        categorical_names: Dictionary where keys are feature columns and values are the categories for the feature

    Raises:
        AttributeError: if categorical_names is not a dict
        AttributeError: if train_data is not a pd.DataFrame

    Example:
        ```python
        >>> explain_data = pd.read_csv('./test/explain_data.csv')
        >>> model = MyModel() # (Trained) model prediction function (has be callable)
        >>> explainer = AnchorsWithImportanceExplainer(model, explain_data)
        >>> explanation = explainer.explain(explain_data.head(10).values) # For the first 10 samples
        >>> explanation.interpret_explanations(n_important_features=2)
        # We can also get the feature importances for the first 10 samples.
        >>> anchorsExtendedExplainer.get_feature_importance(explain_data=explain_data.head(10))
        ```
    """

    def __init__(
        self,
        predict_fn: TP.Callable,
        train_data: pd.DataFrame,
        categorical_names: TP.Dict[str, TP.List] = {},
        disc_perc: TP.Tuple[TP.Union[int, float], ...] = (25, 50, 75),
        *args, **kwargs
    ) -> None:
        if not isinstance(categorical_names, dict):
            raise AttributeError("""
                The attribute categorical_names should be a dictionary
                where the keys are the categorical feature names and the
                values are the categories for each categorical feature.
            """)

        if not isinstance(train_data, pd.DataFrame):
            raise AttributeError("""
                train_data should be a pandas DataFrame.
            """)

        super().__init__(predict_fn, list(train_data.columns), categorical_names)
        self.categorical_names = categorical_names

        super().fit(
            train_data=train_data.values,
            disc_perc=disc_perc,
            *args, **kwargs
        )

    def explain(self,
                X: np.ndarray,
                threshold: float = 0.95,
                delta: float = 0.1,
                tau: float = 0.15,
                batch_size: int = 100,
                coverage_samples: int = 10000,
                beam_size: int = 1,
                stop_on_first: bool = False,
                max_anchor_size: TP.Optional[int] = None,
                min_samples_start: int = 100,
                n_covered_ex: int = 10,
                binary_cache_size: int = 10000,
                cache_margin: int = 1000,
                verbose: bool = False,
                verbose_every: int = 1,
                **kwargs: TP.Any) -> Explanation:
        """
        Explain prediction made by classifier on instance `X`.

        Args:
            X: Instance to be explained.
            threshold: Minimum precision threshold.
            delta: Used to compute `beta`.
            tau: Margin between lower confidence bound and minimum precision or upper bound.
            batch_size: Batch size used for sampling.
            coverage_samples: Number of samples used to estimate coverage from during result search.
            beam_size: The number of anchors extended at each step of new anchors construction.
            stop_on_first: If ``True``, the beam search algorithm will return the
                            first anchor that has satisfies the probability constraint.
            max_anchor_size: Maximum number of features in result.
            min_samples_start: Min number of initial samples.
            n_covered_ex: How many examples where anchors apply to store for each anchor sampled during search
                            (both examples where prediction on samples agrees/disagrees with `desired_label` are stored).
            binary_cache_size: The result search pre-allocates `binary_cache_size` batches for storing the binary arrays
                                returned during sampling.
            cache_margin: When only ``max(cache_margin, batch_size)`` positions in the binary cache remain empty, a new cache
                            of the same size is pre-allocated to continue buffering samples.
            verbose: Display updates during the anchor search iterations.
            verbose_every: Frequency of displayed iterations during anchor search process.

        Returns:
            explanation
                `Explanation` object containing the result explaining the instance with additional metadata as attributes.
                See usage at `AnchorTabular examples`_ for details.
                .. _AnchorTabular examples:
                    https://docs.seldon.io/projects/alibi/en/latest/methods/Anchors.html
        """
        exp = super().explain(
            X=X,
            threshold=threshold,
            delta=delta,
            tau=tau,
            batch_size=batch_size,
            coverage_samples=coverage_samples,
            beam_size=beam_size,
            stop_on_first=stop_on_first,
            max_anchor_size=max_anchor_size,
            min_samples_start=min_samples_start,
            n_covered_ex=n_covered_ex,
            binary_cache_size=binary_cache_size,
            cache_margin=cache_margin,
            verbose=verbose,
            verbose_every=verbose_every
        )

        # This attribute makes pickle serialization crash, so we delete it.
        if hasattr(self, "mab"):
            delattr(self, "mab")

        return exp

    def get_feature_importance(
        self,
        explain_data: pd.DataFrame,
        threshold: float = 0.95,
        print_every: int = 0,
        print_explanations: bool = False,
        n_important_features: int = 3,
        tau: float = 0.15,
        timeout: int = 0) -> AnchorsWithImportanceExplanation:
        """
        Args:
            explain_data:
                Pandas dataframe containing all the instances for which to find an anchor and therefore
                obtain feature importances.
            threshold: To be used in and passed down to the anchor explainer as defined on Alibi's documentation.
                Controls the minimum precision desired when looking for anchors.
                Defaults to 0.95.
            print_every:
                Logging information.
                Defaults to 0 - No logging
            print_explanations:
                Boolean that determines whether to print the explanations at the end of the method or not.
                Defaults to False.
            n_important_features:
                Number of top features that will be printed.
                Defaults to 3.
            tau:
                To be used in and passed down to the anchos explainer as defined on Alibi's documentation.
                Used within the multi-armed bandit part of the optimisation problem.
                Defaults to 0.15
            timeout:
                Maximum time to be spent looking for an Anchor in seconds. A value of 0 means that no timeout
                is set.
                Defaults to 0.

        Returns:
            A list containing all the explanations.
        """

        explanations = []
        if print_every > 0:
            print('Looking for a total of {} explanations'.format(
                len(explain_data))
            )
        for explain_datum_idx, explain_datum in explain_data.iterrows():
            try:
                explanation = run_until_timeout(timeout,
                                  self.explain,
                                  explain_datum.values,
                                  threshold=threshold,
                                  tau=tau)
                explanations.append(explanation)
            except Exception:
                if print_every > 0:
                    print('No anchor found for observation {}'
                                                .format(explain_datum_idx))
                explanations.append('No explanation')

            # Unset timeout
            signal.alarm(0)

            if print_every > 0:
                if len(explanations) % print_every == 0:
                    print(
                        ("""A total of {} observations have been processed """ +
                            """for explaining""").format(len(explanations)))
                    print("{} anchors have already been found".format(
                        sum([1 for explan in explanations
                            if not isinstance(explan, str)])
                    ))
        # Here we have a list with all the anchors explanations that we've been able to find.
        anchorsExtendedExplanation = AnchorsWithImportanceExplanation(
            explain_data=explain_data,
            explanations=explanations,
            categorical=self.categorical_names
        )
        if print_explanations:
            anchorsExtendedExplanation.interpret_explanations(
                n_important_features=n_important_features
            )
        return anchorsExtendedExplanation

    def translate(self, explanation: Explanation) -> str:
        """
        Translates an explanation into simple words

        Args:
            explanation: Alibi explanation object

        """
        coverage = explanation['data']['coverage']
        if type(explanation['data']['precision']) is np.ndarray:
            precision = explanation['data']['precision'][0]
        else:
            precision = explanation['data']['precision']

        if coverage * precision < 0.1:
            quality = "POOR"
        elif 0.1 <= coverage * precision < 0.4:
            quality = "GOOD"
        else:
            quality = "GREAT"

        return "[{} explanation] This anchor explains a {}% of all records of its class with {}% confidence.".format(
            quality,
            round(100 * coverage, 2),
            round(100 * precision, 2)
        )
    
    def save(self, filename):
        """Overwrite to ensure that we use MercuryExplainer.save"""
        MercuryExplainer.save(self, filename=filename)



import time
import pandas as pd
import numpy as np
import typing as TP

from .explainer import MercuryExplainer
from mercury.explainability.explanations.partial_dependence import PartialDependenceExplanation


class PartialDependenceExplainer(MercuryExplainer):
    """
    This explainer will calculate the partial dependences for a ML model.
    Also contains a distributed (pyspark) implementation which
    allows PySpark transformers/pipelines to be explained via PD.

    Args:
        predict_fn (Callable):
            Inference function. This function will take a DataFrame (Pandas
            or PySpark) and must return the output of your estimator (usually a NumPy array
            for a 'plain' python model or a DataFrame in case of PySpark).
        output_col (str):
            Name of the output column name  of the PySpark transformer model.
            This will only be used in case the passed explain data is of type
            pyspark.DataFrame
        max_categorical_thresh (int):
            If a column contains less unique values than this threshold, it will be
            auto-considered categorical.
        quantiles (float or tuple):
            Calculate the quantiles of the model predictions. If type==float,
            (quantiles, 1-quantiles) range will be calculated. If type==tuple, range
            will be `quantiles`. If None, the quantile calculation will be disabled
            (useful for saving time).
        resolution (int):
            Number of different values to test for each non-categorical variable. Lowering this
            value will increase speed but reduce resolution in the plots.
        verbose (bool):
            Print progress status. Default is False.

    Example:
        ```python
        # "Plain python" example
        >>> features = dataset.loc[:, FEATURE_NAMES] # DataFrame with only features
        # You can create a custom inference function.
        >>> def my_inference_fn(feats):
        ...     return my_fitted_model.predict(feats)
        # You can also pass anything as long as it is callable and receives the predictors (e.g. my_fitted_model.predict / predict_proba).
        >>> explainer = PartialDependenceExplainer(my_fitted_model.predict)
        >>> explanation = explainer.explain(features)
        # Plot a summary of partial dependences for all the features.
        >>> explanation.plot()
        # Plot the partial dependence of a single variable.
        >>> fig, ax = plt.subplots()
        >>> explanation.plot_single('FEATURE_1', ax=ax)

        # Pyspark Example
        # Dataset with ONLY feature columns. (Assumed an already trained model)
        >>> features = dataset.drop('target')
        # Inference function
        >>> def my_pred_fn(data):
        ...     temp_df = assembler.transform(data)
        ...     return my_transformer.transform(temp_df)
        # We can tell the explainer not to calculate  partial dependences for certain features. This will save time.
        >>> features_to_ignore = ['FEATURE_4', 'FEATURE_88']
        # We pass our custom inference function and also, tell the explainer which column will hold the transformer's output (necessary only when explaining pyspark models).
        >>> explainer = PartialDependenceExplainer(my_pred_fn, output_col='probability')
        # Explain the model ignoring features.
        >>> explanation = explainer.explain(features, ignore_feats=features_to_ignore)
        ```
    """
    def __init__(self,
            predict_fn: TP.Callable[[TP.Union["pandas.DataFrame", "pyspark.sql.DataFrame"]],  # noqa: F821
                                 TP.Union[np.ndarray, "pyspark.sql.DataFrame"]],  # noqa: F821
            output_col:str = 'prediction',
            max_categorical_thresh:int = 5,
            quantiles:TP.Union[float ,tuple] = 0.05,
            resolution: int = 50,
            verbose:bool = False):
        self.predict_fn = predict_fn
        self.output_col = output_col
        # If the number of different values of a given feature is less
        # than this, we will treat it as categorical
        self._max_categorical_values = max_categorical_thresh
        self.verbose = verbose
        self.resolution = resolution

        if type(quantiles) == float:
            self.compute_quantiles = (quantiles, 1 - quantiles)
        else:
            self.compute_quantiles = quantiles

    def explain(self,
            features: TP.Union["pandas.DataFrame", "pyspark.sql.DataFrame"],  # noqa: F821
            ignore_feats: TP.List[str] = None,
            categoricals: TP.List[str] = None)->PartialDependenceExplanation:
        """
        This method will compute the partial dependences for a ML model.
        This explainer also contains a distributed (pyspark) implementation which
        allows PySpark transformers/pipelines to be explained via PD.

        Args:
            features (pandas.DataFrame or pyspark.sql.DataFrame):
                DataFrame with only the features needed by the model.
                This dataframe should ONLY contain the features consumed by the model and,
                in the PySpark case, the vector-assembled column should be generated inside
                of the predict_fn.
            ignore_feats (List):
                Feature names which won't be explained
            categoricals (List):
                of feature names that will be forced to be taken as
                categoricals. If it's empty, the explainer will guess what columns are
                categorical

        Returns:
            PartialDependenceExplanation containing the explanation results.
        """
        ignore_feats = ignore_feats if ignore_feats else []
        categoricals = categoricals if categoricals else []

        partial_dep_impl = self.__base_impl
        categoricals = set(categoricals)
        if type(features) != pd.DataFrame:
            partial_dep_impl = self.__pyspark_impl

        feat_names = [f for f in list(features.columns) if f not in set(ignore_feats)]

        dependences = partial_dep_impl(features, feat_names, categoricals)

        return PartialDependenceExplanation(dependences)

    def __base_impl(self, features:pd.DataFrame, feat_names:list, categoricals:set)->dict:
        """
        Contains the logic for the base implementation, i.e. the  one
        which uses np.ndarrays/pd.DataFrames and undistributed models.

        Args:
            features (pd.DataFrame): pandas.DataFrame with ONLY the predictors.
            feat_names (list[str]):
                List containing the names of the features that will be explained
            categoricals (set[str]):
                Set with the feature names that will be forced to be categorical.

        Returns:
            Dictionary with the partial dependences of each selected feature
            and whether that feature is categorical or not.
        """
        data = {}
        for colname in feat_names:
            start = time.time()
            data[colname] = {}
            uniques = features[colname].unique()
            nb_uniques = len(uniques)

            if features.loc[:,colname].dtype == np.dtype('O') or colname in categoricals:
                is_categorical = True
            else:
                is_categorical = False

            if not is_categorical and type(features.loc[:,colname].iloc[0].item()) == float and nb_uniques > self._max_categorical_values:
                grid = np.linspace(features[colname].min(), features[colname].max())
            elif not is_categorical and nb_uniques > self._max_categorical_values and type(features.loc[:,colname].iloc[0].item()) == int:
                step_size = max(1, np.abs((features[colname].max() - features[colname].min()) // self.resolution))
                grid = np.arange(features[colname].min(), features[colname].max(), step=step_size)
            else:
                is_categorical = True
                grid = uniques

            if self.verbose:
                type_msg = 'categorical' if is_categorical else 'continuous'
                print(f"Starting calculation for feature {colname} - Type: {type_msg}")
                if is_categorical:
                    print(f"\tValues to test: {grid}")
                else:
                    print(f"\tValue range to test: [{grid[0]}, {grid[-1]}] (only on {self.resolution} items)")

            data[colname]['values'] = grid
            pdep_means, pdep_lquant, pdep_uquant = self.__partial_dep_base(colname, features, grid)
            data[colname]['preds'] = pdep_means
            data[colname]['lower_quantile'] = pdep_lquant
            data[colname]['upper_quantile'] = pdep_uquant
            data[colname]['categorical'] = is_categorical

            elapsed = time.time() - start
            if self.verbose:
                print(f"Partial dependence for feature {colname} calculated. Took {elapsed:.2f} seconds")

        return data

    def __pyspark_impl(self, features: "pyspark.DataFrame", feat_names: list, categoricals: set) -> dict:  # noqa: F821
        """ Contains the logic for the pyspark implementation, i.e. the  one
        which uses pyspark.DataFrames and undistributed models.

        Args:
            features: pyspark.DataFrame with ONLY the predictors.
            feat_names: list[str]
                List containing the names of the features that will be explained
            categoricals: set[str]
                Set with the feature names that will be forced to be categorical.

        Returns:
            Dictionary with the partial dependences of each selected feature
            and whether that feature is categorical or not.
        """
        from pyspark.sql.functions import col
        from pyspark.sql import functions as sqlf

        from pyspark.sql.types import (
            FloatType,
            DoubleType,
            DecimalType,
            LongType,
            IntegerType,
            ShortType,
            ByteType
        )

        # Pyspark names of float and integer datatypes
        float_pyspark_types = (FloatType, DoubleType, DecimalType)
        int_pyspark_types = (FloatType, DoubleType, LongType, IntegerType, ShortType, ByteType)

        data = {}

        # Calculate unique values for disjoint set of columns from categoricals
        cols_to_check = [n for n in feat_names if n not in categoricals]
        # Calculate all statistics at once to save collects
        num_distinct_values = features.agg(
            *(sqlf.countDistinct(col(c)).alias(c) for c in cols_to_check)
        ).collect()[0]
        minimums = features.select([sqlf.min(c).alias(c) for c in features.columns]).collect()[0]
        maximums = features.select([sqlf.max(c).alias(c) for c in features.columns]).collect()[0]

        for colname in feat_names:
            start = time.time()

            data[colname] = {}
            # Determine whether this column is discrete, real or categorical
            is_categorical = False
            if isinstance(features.schema[colname].dataType, float_pyspark_types) and colname not in categoricals:
                grid = np.linspace(minimums[colname], maximums[colname])
            elif isinstance(features.schema[colname].dataType, int_pyspark_types) \
                    and num_distinct_values[colname] > self._max_categorical_values:
                step_size = max(
                    1, np.abs((maximums[colname] - minimums[colname]) // self.resolution)
                )
                grid = np.arange(minimums[colname], maximums[colname], step=step_size)
            else:
                grid = features.select(colname).distinct().rdd.flatMap(lambda x: x).collect()
                is_categorical=True

            if self.verbose:
                type_msg = 'categorical' if is_categorical else 'continuous'
                print(f"Starting calculation for feature {colname} - Type: {type_msg}")
                if is_categorical:
                    print(f"\tValues to test: {grid}")
                else:
                    print(f"\tValue range to test: [{grid[0]}, {grid[-1]}] (only on {self.resolution} items)")

            data[colname]['values'] = grid
            pdep_means, pdep_lquant, pdep_uquant = self.__partial_dep_pyspark(colname, features, grid, is_categorical)
            data[colname]['preds'] = pdep_means
            data[colname]['lower_quantile'] = pdep_lquant
            data[colname]['upper_quantile'] = pdep_uquant
            data[colname]['categorical'] = is_categorical

            elapsed = time.time() - start
            if self.verbose:
                print(f"Partial dependence for feature {colname} calculated. Took {elapsed:.2f} seconds")

        return data

    def __partial_dep_base(self,
            feat_name:str,
            data: "pandas.DataFrame",  # noqa: F821
            grid: TP.Union[list, "np.array"])->"np.ndarray":
        """ Logic for computing the partial dependence of one feature for the base
        implementation (undistributed models)
        """
        preds = []
        lower_quantiles = []
        upper_quantiles = []
        data = data.copy()
        for val in grid:
            data[feat_name] = val
            model_preds = self.predict_fn(data)
            preds.append(np.mean(model_preds, axis=0))
            if self.compute_quantiles:
                lower_quantiles.append(np.quantile(model_preds, q=self.compute_quantiles[0],axis=0))
                upper_quantiles.append(np.quantile(model_preds, q=self.compute_quantiles[1],axis=0))
        return np.array(preds), np.array(lower_quantiles), np.array(upper_quantiles)

    def __partial_dep_pyspark(self,
            feat_name:str,
            data: TP.Union["pandas.DataFrame", "pyspark.sql.DataFrame"],  # noqa: F821
            grid: TP.Union[list, "np.array"],
            is_categorical:bool = False)->"np.ndarray":
        """ Helper method for the PySpark implementation (distributed models).
        Computes the partial dependences of one feature given its type (categorical or not)
        """
        from pyspark.ml.stat import Summarizer
        from pyspark.sql.functions import col
        from pyspark.sql import functions as sqlf

        mean_preds = []
        lower_quantiles = []
        upper_quantiles = []

        for v in grid:
            if type(v) == str:
                temp_df = data.withColumn(feat_name, sqlf.lit(v))
            else:
                temp_df = data.withColumn(feat_name, sqlf.lit(float(v)))
            temp_df = self.predict_fn(temp_df)

            if temp_df.schema[self.output_col].simpleString().split(':')[-1] == 'vector':
                # Aggregate all the probability distributions
                mean_preds.append(temp_df.agg(Summarizer.mean(col(self.output_col))).collect()[0][0].values.tolist())

                if self.compute_quantiles:
                    # Tricky to calculate quantiles as Spark doesnt support this for Vector
                    def _to_cols(row):
                        return tuple(row["probability"].toArray().tolist())

                    pred_cols = temp_df.rdd.map(_to_cols).toDF()
                    quants = np.array(
                        pred_cols.approxQuantile(pred_cols.columns, self.compute_quantiles, 0.15)
                    )
                    lower_quantiles.append(quants[:, 0])
                    upper_quantiles.append(quants[:, 1])
            else:
                # Aggregate all the prediction probabilities/targets
                mean_preds.append(temp_df.agg({self.output_col: "mean"}).collect()[0][0])
                if self.compute_quantiles:
                    quants =temp_df.approxQuantile(self.output_col, self.compute_quantiles, 0.15)
                    lower_quantiles.append(quants[0])
                    upper_quantiles.append(quants[1])

        return np.array(mean_preds), np.array(lower_quantiles), np.array(upper_quantiles)



import typing as TP
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import bokeh.plotting as BP
import bokeh.io as BPIO

from bokeh.models import ColorBar, LinearColorMapper
from bokeh.layouts import layout, row, column


class CounterfactualBasicExplanation(object):
    """
    A Panallet explanation.

    Args:
        from_ (np.ndarray):
            Starting point.
        to_ (np.ndarray):
            Found solution.
        p (float):
            Probability of found solution.
        path (np.ndarray):
            Path followed to get to the found solution.
        path_ps (np.ndarray):
            Probabilities of each path step.
        bounds (np.ndarray):
            Feature bounds used when exploring the probability space.
        explored (np.ndarray):
            Points explored but not visited (available only when backtracking
            strategy is used, empty for Simulated Annealing)
        explored_ps (np.ndarray):
            Probabilities of explored points (available only when backtracking
            strategy is used, empty for Simulated Annealing)
        labels (TP.Optional[TP.List[str]]):
            Labels to be used for each point dimension (used when plotting).

    Raises:
        AssertionError: if from_ shape != to_.shape
        AssertionError: if dim(from_) != 1
        AssertionError: if not 0 <= p <= 1
        AssertionError: if path.shape[0] != path_ps.shape[0]
        AssertionError: if bounds.shape[0] != from_.shape[0]
        AssertionError: if explored.shape[0] != explored_ps.shape[0]
        AssertionError: if len(labels) > 0 and len(labels) != bounds.shape[0]
    """
    def __init__(self,
                 from_: 'np.ndarray',
                 to_: 'np.ndarray',
                 p: float,
                 path: 'np.ndarray',
                 path_ps: 'np.ndarray',
                 bounds: 'np.ndarray',
                 explored: 'np.ndarray' = np.array([]),
                 explored_ps: 'np.ndarray' = np.array([]),
                 labels: TP.Optional[TP.List[str]] = []) -> None:
        # Initial/end points
        assert from_.shape == to_.shape and from_.ndim == 1, 'Invalid dimensions'
        self.from_ = from_
        self.to_ = to_

        # Found solution probability
        assert p >= 0 and p <= 1, 'Invalid probability'
        self.p = p

        # Path followed till solution is found
        assert path.shape[0] == path_ps.shape[0], \
            'Invalid shape for path probabilities, got {} but expected {}'.format(path.shape[0], path_ps.shape[0])
        self.path = path
        self.path_ps = path_ps

        # Used bounds in the solution
        assert bounds.shape[0] == self.from_.shape[0], 'Invalid bounds shape'
        self.bounds = bounds

        assert explored.shape[0] == explored_ps.shape[0], \
            'Invalid shape for explored probabilities, got {} but expected {}'.format(explored.shape[0],
                                                                                      explored_ps.shape[0])
        self.explored = explored

        if labels is not None and len(labels) > 0:
            assert len(labels) == self.bounds.shape[0], 'Invalid number of labels'
        self.labels = labels

    def get_changes(self, relative=True) -> 'np.ndarray':
        """
        Returns relative/absolute changes between initial and ending point.

        Args:
            relative (bool):
                True for relative changes, False for absolute changes.

        Returns:
            (np.ndarray) Relative or absolute changes for each feature.
        """

        if relative:
            # Avoid divs by zero
            aux = self.from_.copy()
            aux[aux == 0.] = 1.
            return (self.to_.squeeze() - self.from_.squeeze()) * 100 / (np.sign(aux) * aux.squeeze())
        else:
            return self.to_.squeeze() - self.from_.squeeze()

    @staticmethod
    def plot_butterfly(data,
                       columns,
                       axis,
                       title: str = '',
                       decimals: int = 1,
                       positive_color: str = '#0A5FB4',
                       negative_color: str = '#DA3851',
                       special_color: str = '#48AE64') -> None:  # pragma: no cover
        data_ = data.copy().squeeze()
        num_labels = [(' {:.' + str(decimals) + 'f} ').format(float(x)) for x in data_]
        are_normal = np.isfinite(data_)
        are_special = False == are_normal
        if sum(are_normal):
            imputed_pos = max(0., np.max(data_[are_normal]))
            imputed_neg = min(0., np.min(data_[are_normal]))
        else:
            imputed_pos, imputed_neg = 1., -1.
        data_[np.logical_and(are_special, np.isnan(data_))] = 0.
        data_[np.logical_and(are_special, np.isposinf(data_))] = imputed_pos
        data_[np.logical_and(are_special, np.isneginf(data_))] = imputed_neg
        rec = axis.barh(range(data_.size), data_)
        axis.tick_params(top='off', bottom='on', left='off', right='off', labelleft='off', labelbottom='on')
        for i, (value, label, num_label, is_normal) in enumerate(zip(data_, columns, num_labels, are_normal)):
            if is_normal:
                color = negative_color if value < 0 else positive_color
            else:
                color = special_color
            num_align, label_align = ('left', 'right') if value > 0 else ('right', 'left')
            axis.text(value, i, num_label, ha=num_align, va='center', color=color, size='smaller')
            axis.text(0, i, ' {} '.format(label), ha=label_align, va='center', color=color, size='smaller')
            rec[i].set_color(color)
        axis.set_axis_off()
        axis.set_title(title)
        axis.title.set_text(title)

    def show(self, figsize: TP.Tuple[int, int] = (12, 6), debug: bool = False,
             path: TP.Optional[str] = None, backend='matplotlib') -> None:  # pragma: no cover
        """
        Creates a plot with the explanation.

        Args:
            figsize (tuple):
                Width and height of the figure (inches if matplotlib backend is used,
                pixels for bokeh backend).
            debug (bool):
                Display verbose information (debug mode).
        """

        def _show(from_: 'np.ndarray', to_: 'np.ndarray', backend='matplotlib',
                  path: TP.Optional[str] = None, debug: bool = False) -> None:
            """ Backend specific show method. """

            if backend == 'matplotlib':
                # It seems we can't decouple figure from axes
                fig = plt.figure(figsize=figsize)

                # LIME-like hbars showing relative differences
                ax = plt.subplot2grid((2, 5), (0, 1))
                CounterfactualBasicExplanation.plot_butterfly(
                    self.get_changes(relative=False), self.labels, ax,
                    title='Absolute delta')

                ax = plt.subplot2grid((2, 5), (0, 3))
                CounterfactualBasicExplanation.plot_butterfly(
                    self.get_changes(relative=True), self.labels, ax,
                    title='Relative delta')

                # Probabilities
                ax = plt.subplot2grid((2, 5), (1, 0), colspan=5)
                xs = np.arange(len(self.path_ps))
                ys = self.path_ps
                cax = ax.scatter(xs, ys, c=ys)
                fig.colorbar(cax)
                ax.plot(xs, ys, '--', c='k', linewidth=.2, alpha=.3)
                ax.grid()
                ax.set_title('Visited itinerary')
                ax.set_xlabel('# Iteration')
                ax.set_ylabel('probability')
                plt.tight_layout()

                if path is not None:
                    plt.savefig(path, output='pdf')
                else:
                    plt.show()

            elif backend == 'bokeh':
                # LIME-like hbars showing relative differences
                values = self.get_changes()
                fig1 = BP.figure(plot_width=400, plot_height=300, y_range=self.labels,
                                 x_range=(min(values), max(values)), x_axis_label='Relative change')
                colors = np.where(values <= 0, '#ff0000', '#00ff00')
                fig1.hbar(y=self.labels, height=0.75, right=values, fill_color=colors)

                # LIME-like hbars showing absolute differences
                values = self.get_changes(relative=False)
                fig2 = BP.figure(plot_width=400, plot_height=300, y_range=self.labels,
                                 x_range=(min(values), max(values)), x_axis_label='Absolute change')
                colors = np.where(values <= 0, '#ff0000', '#00ff00')
                fig2.hbar(y=self.labels, height=0.75, right=values, fill_color=colors, line_color=None)

                # Probabilities
                fig3 = BP.figure(plot_width=800, plot_height=200, x_axis_label='Step', y_axis_label='p')
                xs = np.arange(self.path_ps.size)
                ys = self.path_ps
                color_mapper = LinearColorMapper(palette='Viridis256', low=min(self.path_ps),
                                                 high=max(self.path_ps))
                color_bar = ColorBar(color_mapper=color_mapper, location=(0, 0))
                fig3.circle(xs, ys, size=5, fill_color={'field': 'y', 'transform': color_mapper},
                            fill_alpha=.3, line_color=None)
                fig3.add_layout(color_bar, 'left')
                fig3.line(xs, ys, line_dash='dashed', line_alpha=.3, line_width=.2)

                row1 = row([fig1, fig2])
                row2 = row([fig3])
                lyt = column([row1, row2])

                if path is not None:
                    BPIO.export_png(lyt, filename=path)
                else:
                    BPIO.output_notebook(hide_banner=True)
                    BP.show(lyt)

            else:
                raise ValueError('Unsupported backend')

            if debug:
                self.__verbose()

        _show(self.from_, self.to_, debug=debug, path=path, backend=backend)

    def __verbose(self):  # pragma: no cover
        """ Internal debug information. """

        print('Used bounds:')
        for i in range(self.bounds.shape[0]):
            label = self.labels[i] if self.labels else ''
            print('\t[{}] {}: [{}, {}]'.format(i, label, self.bounds[i][0], self.bounds[i][1]))
        print('Starting point: {}'.format(self.from_))
        print('Found solution: {} with probability {}'.format(self.to_, self.p))
        print('Changes:')
        for i in range(self.from_.shape[0]):
            if self.from_[i] != self.to_[i]:
                label = self.labels[i] if self.labels else ''
                print('\t[{}] {}: {} -> {}'.format(i, label, self.from_[i], self.to_[i]))


class CounterfactualWithImportanceExplanation(object):
    """
    Extended Counterfactual Explanations

    Args:
        explain_data:
            A pandas DataFrame containing the observations for which an explanation has to be found.
        explanations:
            A list containing the results of computing the explanations for explain_data.
        categorical:
            A dictionary containing as key the features that are categorical and as value, the possible
            categorical values.
    """

    def __init__(
            self,
            explain_data: pd.DataFrame,
            counterfactuals: TP.List[dict],
            importances: TP.List[TP.Tuple],
            count_diffs: dict,
            count_diffs_norm: dict
        ) -> None:
        self.explain_data = explain_data
        self.counterfactuals = counterfactuals
        self.importances = importances
        self.count_diffs = count_diffs
        self.count_diffs_norm = count_diffs_norm

    def interpret_explanations(self, n_important_features: int = 3) -> str:
        """
        This method prints a report of the important features obtaiend.

        Args:
            n_important_features:
                The number of imporant features that will appear in the report.
                Defaults to 3.
        """

        importances_str = []
        for n in range(n_important_features):
            importance_str = [imp if isinstance(imp, str) else '{:.2f}'.format(imp) for imp in self.importances[n]]
            importances_str.append(importance_str)

        count_diffs_norm_str = []
        for n in range(n_important_features):
            count_diffs_i = list(self.count_diffs_norm.items())[n]
            count_diff_norm_str = '{} {:.2f}'.format(count_diffs_i[0], count_diffs_i[1])
            count_diffs_norm_str.append(count_diff_norm_str)

        interptretation = """The {} most important features and their importance values according to the first metric (amount features change) are: 
    {}.

According to the second metric (times features change), these importances are: 
    {}""".format(
            n_important_features, 
            ' AND '.join([' '.join(imp_str) for imp_str in importances_str]),
            ' AND '.join(count_diffs_norm_str)
        )
        print(interptretation)
        return interptretation




from typing import List
import numpy as np
try:
    from graphviz import Source
    graphviz_available = True
except Exception:
    graphviz_available = False


class ClusteringTreeExplanation():

    """
    Explanation for ClusteringTreeExplainer. Represents a Decision Tree for the explanation of a clustering
    algorithm.
    Using the plot method generates a visualization of the decision tree (requires graphviz package)

    Args:
        tree: the fitted decision tree
        feature_names: the feature names used in the decision tree

    """

    def __init__(
        self,
        tree: "Node",  # noqa: F821
        feature_names: List = None,
    ):
        self.tree = tree
        self.feature_names = feature_names

    def plot(self, filename: str = "tree_explanation", feature_names: List = None, scalers: dict = None):

        """
        Generates a graphviz.Source object representing the decision tree, which can be visualized in a notebook
        or saved in a file.

        Args:
            filename: filename to save if render() method is called over the returned object
            feature_names: the feature names to use. If not specified, the feature names specified in the constructor
                are used.
            scalers: dictionary of scalers. If passed, the tree will show the denormalized value in the split instead
                of the normalized value. The key is the feature name and the scaler must have the `inverse_transform`
                method

        Returns:
            (graphviz.Source): object representing the decision tree.
        """

        feature_names = self.feature_names if feature_names is None else feature_names
        scalers = {} if scalers is None else scalers

        if not graphviz_available:
            raise Exception("Required package is missing. Please install graphviz")

        if self.tree is not None:
            dot_str = ["digraph ClusteringTree {\n"]
            queue = [self.tree]
            nodes = []
            edges = []
            id = 0
            while len(queue) > 0:
                curr = queue.pop(0)
                if curr.is_leaf():
                    label = "%s\nsamples=\%d\nmistakes=\%d" % (str(self._get_node_split_value(curr)), curr.samples, curr.mistakes) # noqa
                else:
                    feature_name = curr.feature if feature_names is None else feature_names[curr.feature]
                    condition = "%s <= %.3f" % (feature_name, self._get_node_split_value(curr, feature_name, scalers))
                    label = "%s\nsamples=\%d" % (condition, curr.samples) # noqa
                    queue.append(curr.left)
                    queue.append(curr.right)
                    edges.append((id, id + len(queue) - 1))
                    edges.append((id, id + len(queue)))
                nodes.append({"id": id,
                              "label": label,
                              "node": curr})
                id += 1
            for node in nodes:
                dot_str.append("n_%d [label=\"%s\"];\n" % (node["id"], node["label"]))
            for edge in edges:
                dot_str.append("n_%d -> n_%d;\n" % (edge[0], edge[1]))
            dot_str.append("}")
            dot_str = "".join(dot_str)
            s = Source(dot_str, filename=filename + '.gv', format="png")
            return s

    def _get_node_split_value(self, node, feature_name=None, scalers=None):
        if (feature_name is not None) and (scalers is not None) and (feature_name in scalers):
            return scalers[feature_name].inverse_transform(np.array([node.value]).reshape(1, -1))[0][0]
        else:
            return node.value



import typing as TP
import numpy as np
import pandas as pd
import heapq
import itertools

class AnchorsWithImportanceExplanation(object):
    """
    Extended Anchors Explanations

    Args:
        explain_data:
            A pandas DataFrame containing the observations for which an explanation has to be found.
        explanations:
            A list containing the results of computing the explanations for explain_data.
        categorical:
            A dictionary containing as key the features that are categorical and as value, the possible
            categorical values.
    """

    def __init__(
            self,
            explain_data: pd.DataFrame,
            explanations: TP.List,
            categorical: dict = {}
        ) -> None:
        self.explain_data = explain_data
        self.explanations = explanations
        self.categorical = categorical
    
    def interpret_explanations(self, n_important_features: int) -> str:
        """
        This method prints a report of the important features obtaiend.

        Args:
            n_important_features:
                The number of imporant features that will appear in the report.
                Defaults to 3.
        """
        names = []
        explanations_found = [explan for explan in self.explanations if not isinstance(explan, str)]
        for expl in explanations_found:
            for name in expl.data['anchor']:
                # split without an argument splits by spaces, and in every item in expl['names']
                # the first word refers to the feature name.
                if (
                    (' = ' in name) or 
                    ((len(self.categorical) > 0) and (name in [item for sublist in list(self.categorical.values()) for item in sublist]))
                    ):
                    names.append(name)
                else:
                    names.append(' '.join(name[::-1].split('.', 1)[1][::-1].split()[:-1]))

        unique_names, count_names = np.unique(names, return_counts=True)
        top_feats = heapq.nlargest(n_important_features, count_names)
        print_values = ['The ', str(n_important_features), ' most common features are: ']
        unique_names_ordered = sorted(unique_names.tolist(), key=lambda x: count_names[unique_names.tolist().index(x)], reverse=True)
        count_names_ordered = sorted(count_names.tolist(), reverse=True)
        n_explanations = 0
        for unique_name, count_name in zip(unique_names_ordered[:n_important_features], count_names_ordered[:n_important_features]):
            if n_explanations == 0:
                print_values.append([unique_name, ' with a frequency of ', 
                    str(count_name), ' (', str(100 * count_name / len(explanations_found)), '%) '])
            elif n_explanations == n_important_features - 1:
                print_values.append([' and ', unique_name, ' with a frequency of ', 
                    str(count_name), ' (', str(100 * count_name / len(explanations_found)), '%) '])
            else:
                print_values.append([', ',unique_name, ' with a frequency of ', 
                    str(count_name), ' (', str(100 * count_name / len(explanations_found)), '%) '])
        n_explanations += 1
        interptretation = ''.join(list(itertools.chain(*print_values)))
        print(interptretation)
        return interptretation



import signal


def run_until_timeout(timeout, fn, *args, **kwargs):
    """
    Timeout function to stop the execution in case it takes longer than the timeout argument.
    After timeout seconds it will raise an exception.

    Args:
        timeout: Number of seconds until the Exception is raised.
        fn: Function to execute.
        *args: args of the function fn.
        **kwargs: keyword args passed to fn.

    Example:
        >>> explanation = run_until_timeout(timeout, explainer.explain, data=data)

    """
    def signal_handler(signum, frame):
        raise Exception('Timeout: explanation took too long...')

    if timeout > 0:
        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(timeout)
    else:
        signal.alarm(0)
    return fn(*args, **kwargs)


from .counter_fact_basic import CounterFactualExplainerBasic
from .shuffle_importance import ShuffleImportanceExplainer
from .partial_dependence import PartialDependenceExplainer
from .clustering_tree_explainer import ClusteringTreeExplainer
from .explainer import MercuryExplainer

# Classes with alibi dependencies
try:
    from .ale import ALEExplainer
    from .anchors import AnchorsWithImportanceExplainer
    from .counter_fact_importance import CounterfactualExplainer, CounterfactualProtoExplainer
except ModuleNotFoundError:
    # if import fails, then import Dummy Class with the same name which raises Error if instantiated
    from ._dummy_alibi_explainers import (
        ALEExplainer,
        AnchorsWithImportanceExplainer,
        CounterfactualExplainer,
        CounterfactualProtoExplainer
    )



from setuptools import setup, Extension
import numpy

setup_args = dict(
    ext_modules = [
        Extension(
            "cut_finder",
            ["mercury/explainability/explainers/_tree_splitters/cut_finder.pyx"],
            extra_compile_args=['-fopenmp'],
            extra_link_args=['-fopenmp'],
            include_dirs=[numpy.get_include()]
        )
    ]
)
setup(**setup_args)


import typing as TP
import numpy as np
import pandas as pd

from .explainer import MercuryExplainer
from .cf_strategies import SimulatedAnnealing, Backtracking

from mercury.explainability.explanations.counter_factual import (
    CounterfactualBasicExplanation
)


class CounterFactualExplainerBasic(MercuryExplainer):
    """
    Explains predictions on tabular (i.e. matrix) data for binary/multiclass classifiers.
    Currently two main strategies are implemented: one following a backtracking strategy and
    another following a probabilistic process (simulated annealing strategy).

    Args:
        train (TP.Union['np.ndarray', pd.DataFrame]):
            Training dataset to extract feature bounds from.
        fn (TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], TP.Union[float, 'np.ndarray']]):
            Classifier `predict_proba`-like function. Note that the returned probabilities
            must be valid, ie. the values must be between 0 and 1.
        labels (TP.List[str]):
            List of labels to be used when plotting results. If DataFrame used, labels take
            dataframe column names. Default is empty list.
        bounds (TP.Optional['np.ndarray']):
            Feature bounds used when no train data is provided (shape must match labels').
            Default is None.
        n_steps (int):
            Parameter used to indicate how small/large steps should be when exploring
            the space (default is 200).

    Raises:
        AssertionError:
            if bounds.size <= 0 when no train data is provided |
            if bounds.ndim != 2 when no train data is provided |
            if bounds.shape[1] != 2 when no train data is provided |
            if bounds.shape[0] != len(labels)
        TypeError:
            if train is not a DataFrame or numpy array.
    """

    def __init__(self,
                 train: TP.Union[TP.Optional['np.ndarray'], TP.Optional[pd.DataFrame]],
                 fn: TP.Callable[[TP.Union['np.ndarray', pd.DataFrame]], TP.Union[float, 'np.ndarray']],
                 labels: TP.List[str] = [],
                 bounds: TP.Optional['np.ndarray'] = None,
                 n_steps: int = 200) -> None:
        if train is None:
            # If data is not provided, labels and bounds are required
            assert bounds.size > 0, 'Bounds are required if no data is provided'
            assert bounds.ndim == 2 and bounds.shape[1] == 2, 'min/max values are required for each feature'
            assert len(labels) == bounds.shape[0], \
                'Labels and bound shapes must match, got {} and {} respectively' \
                    .format(len(labels), bounds.shape[0])
            # min/max values for each feature
            self.labels = labels
            self.bounds = bounds
        else:
            # Compute bounds
            if isinstance(train, pd.DataFrame):
                self.labels = train.columns.tolist()
                self.bounds = train.describe().loc[['min', 'max']].values.T
                assert len(self.labels) == self.bounds.shape[0], \
                    'Labels and bound shapes must match, got {} and {} respectively' \
                        .format(len(self.labels), self.bounds.shape[0])
            elif isinstance(train, np.ndarray):
                self.labels = labels
                self.bounds = np.stack([
                    np.apply_along_axis(np.min, 0, train),
                    np.apply_along_axis(np.max, 0, train)], axis=1)
                assert len(self.labels) == self.bounds.shape[0], \
                    'Labels and bound shapes must match, got {} and {} respectively' \
                        .format(len(self.labels), self.bounds.shape[0])
            else:
                raise TypeError('Invalid type for argument train, got {} but expected numpy array or pandas dataframe'.
                                format(type(train)))

        # Compute steps
        self.n_steps = n_steps
        self.step = (self.bounds[:, 1] - self.bounds[:, 0]) / self.n_steps

        # Function to be evaluated on optimization
        self.fn = fn

    def explain(self,
                from_: 'np.ndarray',
                threshold: float,
                class_idx: int = 1,
                kernel: TP.Optional['np.ndarray'] = None,
                bounds: TP.Optional['np.ndarray'] = None,
                step: TP.Optional['np.ndarray'] = None,
                strategy: str = 'backtracking',
                report: bool = False,
                keep_explored_points: bool = True,
                **kwargs) -> CounterfactualBasicExplanation:
        """
        Roll the panellet down the valley and find an explanation.

        Args:
            from_ ('np.ndarray'):
                Starting point.
            threshold (float):
                Probability to be achieved (if path is found).
            class_idx (int):
                Class to be explained (e.g. 1 for binary classifiers).
            kernel (TP.Optional['np.ndarray']):
                Used to penalize certain dimensions when trying to move around
                the probability space (some dimensions may be more difficult to explain,
                hence don't move along them). Default is np.ones(n), meaning all dimensions
                can be used to move around the space (must be a value between 0 and 1).
            bounds (TP.Optional['np.ndarray']):
                Feature bound values to be used when exploring the probability space. If not
                specified, the ones extracted from the training data are used instead.
            step (TP.Optional['np.ndarray']):
                Step values to be used when moving around the probability space. If not specified,
                training bounds are divided by 200 (arbitrary value) and these are used as step value.
            strategy (str):
                If 'backtracking', the backtracking strategy is used to move around the probability space.
                If 'simanneal', the simulated annealing strategy is used to move around the probability space.
            report (bool):
                Whether to report the algorithm progress during the execution.
            keep_explored_points (bool):
                Whether to keep the points that the algorithm explores. Setting it to False will decrease
                the computation time and memory usage in some cases. Default value is True.

        Raises:
            AssertionError:
                If `from_` number of dimensions is != 1 |
                If `from_` shape does not match `bounds` shape |
                If `bounds` shape is not valid |
                If `step` shape does not match `bounds` shape |
            ValueError:
                if strategy is not 'backtacking' or 'simanneal'.

        Returns:
            explanation (CounterfactualBasicExplanation):
                CounterfactualBasicExplanation with the solution found and how it differs from the starting point.
        """

        if kernel is None:
            self.kernel = np.ones(from_.shape[0])
        else:
            self.kernel = kernel

        assert from_.ndim == 1, \
            'Invalid starting point shape, got {} but expected unidimensional vector'.format(from_.shape)

        if bounds is not None:
            assert from_.shape[0] == bounds.shape[0], \
                'Starting point and bounds shapes should match, got {} and {}'.format(from_.shape, bounds.shape[0])

        # Update bounds based on reference point
        l_bounds = self.bounds.copy()
        for i, bound in enumerate(l_bounds):
            new_min = bound[0]
            new_max = bound[1]
            if from_[i] < bound[0]:
                new_min = from_[i]
            elif from_[i] > bound[1]:
                new_max = from_[1]
            l_bounds[i, 0] = new_min
            l_bounds[i, 1] = new_max

        # Update bounds, if new_bounds are specified
        if bounds is not None:
            assert bounds.shape == l_bounds.shape, \
                'Invalid dimensions for new bounds, got {} but expected {}'.format(bounds.shape, l_bounds.shape)
            for i, bound in enumerate(bounds):
                # Update bound only if starting point is within it in this dimension
                if bound[0] <= from_[i] and bound[1] >= from_[i]:
                    l_bounds[i] = bounds[i]

        if step is not None:
            assert step.shape[0] == l_bounds.shape[0], \
                'Invalid step shape, got {} but expected {}'.format(step.shape, l_bounds.shape[0])
            self.step = step

        if strategy == 'backtracking':
            # Backtracking strategy
            sol, p, visited, explored = Backtracking(from_, l_bounds, self.step, self.fn, class_idx,
                                                           threshold=threshold, kernel=self.kernel, report=report,
                                                           keep_explored_points=keep_explored_points).run(
                **kwargs)
            ps = visited[:, -1]
            visited = visited[:, :-1]
            if keep_explored_points:
                explored_points = explored[:, :-1]
                explored_ps = explored[:, -1]
            else:
                explored_points = np.array([])
                explored_ps = np.array([])
            return CounterfactualBasicExplanation(
                from_, sol, p, visited, ps, l_bounds, explored_points,
                explored_ps, labels=self.labels)
        elif strategy == 'simanneal':
            # Simulated Annealing strategy
            sol, p, visited, energies = SimulatedAnnealing(from_, l_bounds, self.step, self.fn, class_idx,
                                                                 threshold=threshold, kernel=self.kernel,
                                                                 report=report).run(**kwargs)
            return CounterfactualBasicExplanation(from_, sol, abs(p), visited, energies[:-1], l_bounds,
                                                   labels=self.labels)
        else:
            raise ValueError('Invalid strategy')
